package commandline;

import gui.FSMViewer;

import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;

import parser.FSMParseException;
import parser.FSMParser;
import core.FiniteStateMachine;

/**
 * This class provides command-line access to the functions of FSMCalc
 * 
 * @author Leopold Haller
 * 
 */
public class FSMMain {
	private static final String STDIN_INPUT = "!stdin";

	private static final String ARG_LTS = "-lts";

	private static final String ARG_FSM = "-fsm";

	private static final String CMD_ACCEPTS = "acc";

	private static final String CMD_ACCEPTS_FULL = "accepts";

	private static final String CMD_VISUALIZE = "vis";

	private static final String CMD_VISUALIZE_FULL = "visualize";

	private static final String CMD_MINIMIZE = "min";

	private static final String CMD_MINIMIZE_FULL = "minimize";

	private static final String CMD_PRODUCT = "prod";

	private static final String CMD_PRODUCT_FULL = "product";

	private static final String CMD_PARALLEL = "par";

	private static final String CMD_PARALLEL_FULL = "parallel";

	private static final String CMD_POWER = "pow";

	private static final String CMD_POWER_FULL = "power";

	private static final String CMD_LINK = "lnk";

	private static final String CMD_LINK_FULL = "link";
	
	private static final String CMD_PARTIALORDER = "po";
	
	private static final String CMD_PARTIALORDER_FULL = "partialorder";
	
	private static final String CMD_PARTIALORDER_CHECKER = "poc";
	
	private static final String CMD_PARTIALORDER_CHECKER_FULL = "partialorderchecker";

	private static final String CMD_COMPLEMENT = "comp";

	private static final String CMD_COMPLEMENT_FULL = "complement";
	
	private static final String CMD_INVARIANT = "mkinv";

	private static final String CMD_INVARIANT_FULL = "makeinvariant";
	
	private static final String CMD_CHECK = "chk";

	private static final String CMD_CHECK_FULL = "check";
	
	private static final String CMD_HELP = "-h";

	private static final String CMD_HELP_FULL = "--help";
	
    private static final String OUTPUT_NORMAL = "normal";
    
	private static final String OUTPUT_OFF = "off";
	
	private static final String OUTPUT_VERBOSE = "verbose";
	
	private static FiniteStateMachine stdIn = null;

	public static void main(String[] args) {
		
		if (args.length == 0) {
			printUnknown();
			return;
		}
		
		String outputMode = System.getProperty("output");
		
		long startTime = System.nanoTime();
		
		if((outputMode == null) || (outputMode.equalsIgnoreCase(OUTPUT_NORMAL))) {
			FiniteStateMachine.setOutputMode(FiniteStateMachine.OUTPUT_NORMAL);
		} else if(outputMode.equalsIgnoreCase(OUTPUT_OFF)) {
			FiniteStateMachine.setOutputMode(FiniteStateMachine.OUTPUT_OFF);
		} else if(outputMode.equalsIgnoreCase(OUTPUT_VERBOSE)) {
			FiniteStateMachine.setOutputMode(FiniteStateMachine.OUTPUT_VERBOSE);
		} else {
		    FiniteStateMachine.setOutputMode(FiniteStateMachine.OUTPUT_NORMAL);
        }
				
		
		if (args[0].equalsIgnoreCase(CMD_ACCEPTS)
				|| args[0].equalsIgnoreCase(CMD_ACCEPTS_FULL)) {
			
			FiniteStateMachine inpFSM = readInputFSM(1, args);
			
			String[] inputSymbols = new String[args.length - 3];
			for (int i = 0; i < inputSymbols.length; i++) {
				inputSymbols[i] = args[i + 3];
			}

			System.out.println(inpFSM.accepts(inputSymbols));

		} else if (args[0].equalsIgnoreCase(CMD_LINK)
				|| args[0].equalsIgnoreCase(CMD_LINK_FULL)) {
			if (args.length < 5) {
				System.err.println("syntax error, see usage");
				System.exit(1);
			}
			
			FiniteStateMachine inpFSM = readInputFSM(1, args);
			
			String[] linkings;
			
			if( (args.length&1) == 0) {
				linkings = new String[args.length - 4];
				System.arraycopy(args, 3, linkings, 0, linkings.length);
			} else {
				linkings = new String[args.length - 3];
				System.arraycopy(args, 3, linkings, 0, linkings.length);
			}
						
			FiniteStateMachine result = inpFSM.link(linkings);

			if ((args.length&1) == 1) {
				System.out.print(result.toString());
			} else {
				writeResult(result, args[args.length-1]);
			}

		} else if (args[0].equalsIgnoreCase(CMD_VISUALIZE)
				|| args[0].equalsIgnoreCase(CMD_VISUALIZE_FULL)) {
			if (args.length < 3) {
				System.err.println("syntax error, see usage");
				System.exit(1);
			}

			FiniteStateMachine inpFSM = readInputFSM(1, args);
			
			try {
				new FSMViewer(inpFSM, args[2]);	
			} catch(NoClassDefFoundError e) {
				System.err.println("Some class definitions could not be found. Please make sure the JGraph library is on the classpath (www.jgraph.com). \nThe following error was encountered: \n" + e.getMessage());
				System.exit(1);
			}
				
			
			
			
		} else if (args[0].equalsIgnoreCase(CMD_MINIMIZE)
				|| args[0].equalsIgnoreCase(CMD_MINIMIZE_FULL)) {
			if (args.length < 3) {
				System.err.println("syntax error, see usage");
				System.exit(1);
			}
			FiniteStateMachine inpFSM = readInputFSM(1, args);
			FiniteStateMachine result = inpFSM.minimize();

			if (args.length <= 3) {
				System.out.print(result.toString());
			} else {
				writeResult(result, args[3]);
			}
			
			printComparision(result, inpFSM);
			
		} else if (args[0].equalsIgnoreCase(CMD_POWER)
				|| args[0].equalsIgnoreCase(CMD_POWER_FULL)) {
			if (args.length < 3) {
				System.err.println("syntax error, see usage");
				System.exit(1);
			}
			FiniteStateMachine inpFSM = readInputFSM(1, args);
			FiniteStateMachine result = inpFSM.power();

			if (args.length <= 3) {
				System.out.print(result.toString());
			} else {
				writeResult(result, args[3]);
			}
			printComparision(result, inpFSM);
		} else if (args[0].equalsIgnoreCase(CMD_PRODUCT)
				|| args[0].equalsIgnoreCase(CMD_PRODUCT_FULL)) {
			if (args.length < 5) {
				System.err.println("syntax error, see usage");
				System.exit(1);
			}

			FiniteStateMachine inpFSM1 = readInputFSM(1, args);
			FiniteStateMachine inpFSM2 = readInputFSM(3, args);
					
			FiniteStateMachine result = inpFSM1.product(inpFSM2);

			if (args.length <= 5) {
				System.out.print(result.toString());
			} else {
				writeResult(result, args[5]);
			}
			printComparision(result, inpFSM1, inpFSM2);
		} else if (args[0].equalsIgnoreCase(CMD_PARALLEL)
				|| args[0].equalsIgnoreCase(CMD_PARALLEL_FULL)) {
			if (args.length < 5) {
				System.err.println("syntax error, see usage");
				System.exit(1);
			}

			FiniteStateMachine inpFSM1 = readInputFSM(1, args);
			FiniteStateMachine inpFSM2 = readInputFSM(3, args);
			FiniteStateMachine result = inpFSM1.parallel(inpFSM2);
			
			if (args.length <= 5) {
				System.out.print(result.toString());
			} else {
				writeResult(result, args[5]);
			}
			
			printComparision(result, inpFSM1, inpFSM2);
		
		} else if (args[0].equalsIgnoreCase(CMD_COMPLEMENT)
				|| args[0].equalsIgnoreCase(CMD_COMPLEMENT_FULL)) {
			if (args.length < 3) {
				System.err.println("syntax error, see usage");
				System.exit(1);
			}

			FiniteStateMachine inpFSM = readInputFSM(1, args);
			FiniteStateMachine result = inpFSM.complementLanguage();
			
			if (args.length <= 3) {
				System.out.print(result.toString());
			} else {
				writeResult(result, args[3]);
			}
			
			printComparision(result, inpFSM);
		} else if(args[0].equalsIgnoreCase(CMD_PARTIALORDER)
				||args[0].equalsIgnoreCase(CMD_PARTIALORDER_FULL)) {
			
			if(args.length < 5) {
				System.err.println("syntax error, see usage");
				System.exit(1);
			}
			
			FiniteStateMachine inputs[] = new FiniteStateMachine[ (args.length - 1) / 2];
			for(int i = 0; i < inputs.length; i++) {
				inputs[i] = readInputFSM(2*i + 1, args); 
			}
			
			FiniteStateMachine result = FiniteStateMachine.partialOrderReduction(null, inputs);
			
			if( (args.length & 1) == 0) {
				writeResult(result, args[args.length-1]);
			} else {
				System.out.println(result.toString());
			}
								
			printComparision(result, inputs);
			
		} else if(args[0].equalsIgnoreCase(CMD_PARTIALORDER_CHECKER)
				||args[0].equalsIgnoreCase(CMD_PARTIALORDER_CHECKER_FULL)) {
			
			if(args.length < 6) {
				System.err.println("syntax error, see usage");
				System.exit(1);
			}
			
			FiniteStateMachine checker;
			FiniteStateMachine inputs[] = new FiniteStateMachine[ (args.length - 3) / 2];
			
			checker = readInputFSM(1, args);
			
			for(int i = 0; i < inputs.length; i++) {
				inputs[i] = readInputFSM(2*i + 3, args); 
			}
			
			FiniteStateMachine result = FiniteStateMachine.partialOrderReduction(checker, inputs);
			
			if( (args.length & 1) == 0) {
				writeResult(result, args[args.length-1]);
			} else {
				System.out.println(result.toString());
			}
			
			FiniteStateMachine inputsWithChecker[] = new FiniteStateMachine[inputs.length + 1];
			System.arraycopy(inputs, 0, inputsWithChecker, 1, inputs.length);
			inputsWithChecker[0] = checker;
			printComparision(result, inputsWithChecker);
						
		} else if(args[0].equalsIgnoreCase(CMD_INVARIANT)
				||args[0].equalsIgnoreCase(CMD_INVARIANT_FULL)) {
			if(args.length < 4) {
				System.err.println("syntax error, see usage");
				System.exit(1);
			}
			
			FiniteStateMachine checker;
			FiniteStateMachine inputs[] = new FiniteStateMachine[ (args.length - 3) / 2];
			
			checker = readInputFSM(1, args);
			
			for(int i = 0; i < inputs.length; i++) {
				inputs[i] = readInputFSM(2*i + 3, args); 
			}
			
			FiniteStateMachine result = checker.makeInvariantTo(inputs);
			
			if( (args.length & 1) == 0) {
				writeResult(result, args[args.length-1]);
			} else {
				System.out.println(result.toString());
			}
						
			
		} else if(args[0].equalsIgnoreCase(CMD_CHECK)
				||args[0].equalsIgnoreCase(CMD_CHECK_FULL)) {
			FiniteStateMachine fsm = readInputFSM(1, args);
			System.out.println(fsm.hasReachableFinalState());
			
		} else if(args[0].equalsIgnoreCase(CMD_HELP) 
				||args[0].equalsIgnoreCase(CMD_HELP_FULL)){
			
			printUsage();
			System.exit(0);
		} else {
			printUnknown();
			System.exit(0);
		}
		
		if(FiniteStateMachine.getOutputMode() != FiniteStateMachine.OUTPUT_OFF) 
			System.err.println("finished after " + (System.nanoTime() - startTime)/1000000000d + " seconds");

	}

	private static void printUnknown() {
		System.err.println("Unknown command, call with -h or --help for help");
	}
	
	private static void printUsage() {
		System.out.println("Usage: ");
		
		System.out.println("java -jar fsmcalc.jar <operation>");
		System.out.println("operations:");
		System.out.println("  (" + CMD_ACCEPTS + "|" + CMD_ACCEPTS_FULL + ") <type> <inputfile> {<input> }\n");
		System.out.println("  (" + CMD_VISUALIZE + "|" + CMD_VISUALIZE_FULL + ") <type> <inputfile>\n");
		System.out.println("  (" + CMD_MINIMIZE + "|" + CMD_MINIMIZE_FULL + ") <type> <inputfile> [<outputfile>]\n");
		System.out.println("  (" + CMD_POWER + "|" + CMD_POWER_FULL + ") <type> <inputfile> [<outputfile>]\n");
		System.out.println("  (" + CMD_PRODUCT + "|" + CMD_PRODUCT_FULL + ") <type> <inputfile1> <type> <inputfile2> [<outputfile>]\n");
		System.out.println("  (" + CMD_PARALLEL + "|" + CMD_PARALLEL_FULL + ") <type> <inputfile1> <type> <inputfile2> [<outputfile>]\n");
		System.out.println("  (" + CMD_COMPLEMENT + "|" + CMD_COMPLEMENT_FULL + ") <type> <inputfile> [<outputfile>]\n");
		System.out.println("  (" + CMD_LINK + "|" + CMD_LINK_FULL + ") <type> <inputfile> <origsymbol>\n  <newsymbol> {<origsymbol> <newsymbol>} [<outputfile>]\n");
		System.out.println("  (" + CMD_PARTIALORDER + "|" + CMD_PARTIALORDER_FULL + ") <type> <inputfile> <type> <inputfile>\n  {<type> <inputfile> } [<outputfile>]\n");
		System.out.println("  (" + CMD_PARTIALORDER_CHECKER + "|" + CMD_PARTIALORDER_CHECKER_FULL + ") <checkertype> <checkerfile> <type> <inputfile>\n  <type> <inputfile> {<type> <inputfile> } [<outputfile>]\n");
		System.out.println("  (" + CMD_INVARIANT + "|" + CMD_INVARIANT_FULL + ") <checkertype> <checkerfile> <type> <inputfile>\n  {<type> <inputfile> } [<outputfile>]\n");
		System.out.println("  (" + CMD_CHECK + "|" + CMD_CHECK_FULL + ") <type> <inputfile>\n");
		System.out.println("\n<type> and <checkertype> are either \"-fsm\" or \"-lts\"");
		System.out.println("<inputfile> and <checkerfile> are either <filename> or \"" + STDIN_INPUT + "\"");
		System.out.println("Examples:\n  java -jar fsmcalc.jar poc -fsm checker.fsm -lts " + STDIN_INPUT + " -lts in1.lts"
				+ "\n  java -jar fsmcalc.jar acc -fsm test.fsm a b a k u s");
		
		
	}

	private static FiniteStateMachine readInputFSM(int argsindex, String[] args) {
		boolean lts;

		if (args[argsindex].equalsIgnoreCase(ARG_LTS)) {
			lts = true;
		} else if (args[argsindex].equalsIgnoreCase(ARG_FSM)) {
			lts = false;
		} else {
			System.err.println("expected \"-lts\" or \"-fsm\"");
			System.exit(1);
			return null;
		}

		if (argsindex + 1 >= args.length) {
			System.err.println("expected filename");
			System.exit(1);

		}
		
		FiniteStateMachine fsm;
		
		if(args[argsindex + 1].equals(STDIN_INPUT)) {
			fsm = readFromStdIn();
		} else {
			fsm = readInputFSM(args[argsindex + 1]);	
		}

		if(fsm.getInitialStates().size() == 0) {
			System.err.println("WARNING: " + "fsm read from " + args[argsindex + 1] + " contains no initial states");
		}
		
		if (lts) {
			return fsm.makeAllStatesFinal();
		} else {
			return fsm;
		}

	}

	private static FiniteStateMachine readFromStdIn() {
		if(stdIn != null)
			return stdIn;
		
						
		try {
			if(FiniteStateMachine.getOutputMode() != FiniteStateMachine.OUTPUT_OFF)
				System.err.println("reading from stdin...");
			stdIn = FSMParser.parseFSM(new InputStreamReader(System.in));
		} catch (IOException e) {
			System.err.println("IO error ocurred while trying to read stdin: ");
			System.err.println(e.getMessage());
			System.exit(1);
		} catch (FSMParseException e) {
			System.err.println("parse error ocurred while reading stdin: ");
			System.err.println(e.getMessage());
			System.exit(2);
		}
		
		
		return stdIn;
	}
		
	
	private static FiniteStateMachine readInputFSM(String filename) {
		FiniteStateMachine fsm = null;
		try {
			fsm = FSMParser.readFSM(filename);
		} catch (IOException e) {
			System.err.println("IO error ocurred while trying to read file \""
					+ filename + "\":");
			System.err.println(e.getMessage());
			System.exit(1);
		} catch (FSMParseException e) {
			System.err.println("parse error ocurred in file \"" + filename
					+ "\":");
			System.err.println(e.getMessage());
			System.exit(2);
		}

		return fsm;
	}

	private static void writeResult(FiniteStateMachine fsm, String filename) {
		if(FiniteStateMachine.getOutputMode() != FiniteStateMachine.OUTPUT_OFF)
			System.err.println("writing result to file " + filename);
		try {
 			FileWriter fw = new FileWriter(filename);
			fw.write(fsm.toString());
			fw.close();
		} catch (IOException e) {
			System.err.println("IO error ocurred while trying to write to result file \""
							+ filename + "\":");
			System.err.println(e.getMessage());
			System.exit(1);
		}

	}

	private static void printComparision(FiniteStateMachine result,
			FiniteStateMachine... original) {
		
		if(FiniteStateMachine.getOutputMode() == FiniteStateMachine.OUTPUT_OFF) 
			return;
		
		System.err.print("         \tresult");
		for (int i = 0; i < original.length; i++) {
			System.err.print("\tin" + i);
		}

		System.err.println();

		System.err.print("states:  \t" + result.getStates().size());
		for (FiniteStateMachine fsm : original)
			System.err.print("\t" + fsm.getStates().size());

		System.err.println();

		System.err.print("initial:\t" + result.getInitialStates().size());
		for (FiniteStateMachine fsm : original)
			System.err.print("\t" + fsm.getInitialStates().size());

		System.err.println();

		System.err.print("final:   \t" + result.getFinalStates().size());
		for (FiniteStateMachine fsm : original)
			System.err.print("\t" + fsm.getFinalStates().size());

		System.err.println();

	}
}

