package core;

import gui.FSMViewer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.Stack;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Map.Entry;

/**
 * This class represents a finite state machine (E,S,I,T,F) where
 * <ul>
 * <li>E is the input alphabet (getAlphabet(), represented implicitely in the
 * state class' transition maps)
 * <li>S is a finite non-empty set of states (getStates())
 * <li>I is the set of initial states (getInitialStates())
 * <li>T is the state transition function (represented in the state class'
 * transition maps)
 * <li>F is the set of final states (getFinalStates())
 * </ul>
 * The class is very similar to the mathematical construct. The set of states S, I and F
 * are represented Collections. The transitions are not globally defined in a transition relation, 
 * but locally in the transitioning states themselves. The alphabet is implicitely defined through the
 * used transition symbols.
 * 
 * @author Leopold Haller
 */

public class FiniteStateMachine {
	public static final int OUTPUT_NORMAL = 1;

	public static final int OUTPUT_OFF = 0;

	public static final int OUTPUT_VERBOSE = 2;

	private static int outputMode = 0;
	
	private int anonCounter = 0;

	private TreeSet<State> finalState = new TreeSet<State>();

	private TreeSet<State> initialState = new TreeSet<State>();

	private TreeMap<String, State> stateMap = new TreeMap<String, State>();

	private Collection<State> state = stateMap.values();

	// Create StateTuples representing all combinations of the given states
	private static Collection<StateTuple> createInitialStateTuples(
			FiniteStateMachine fsm[], FiniteStateMachine checker) {

		int initialStateCount = 1;

		for (int i = 0; i < fsm.length; i++) {
			initialStateCount *= fsm[i].getInitialStates().size();
		}

		ArrayList<StateTuple> initStates = new ArrayList<StateTuple>(
				initialStateCount);
		createInitialStateTuples_aux(0, new State[0], fsm, checker, initStates);

		return initStates;
	}

	private static void createInitialStateTuples_aux(int recursionDepth,
			State stateArray[], FiniteStateMachine fsm[],
			FiniteStateMachine checker, Collection<StateTuple> initStates) {

		if (recursionDepth >= fsm.length) {
			if (checker != null) {
				for (State checkerInit : checker.getInitialStates()) {
					StateTuple tuple = new StateTuple(stateArray, checkerInit);
					initStates.add(tuple);
				}
			} else {
				initStates.add(new StateTuple(stateArray, null));
			}
			return;
		}

		Collection<State> initialStates = fsm[recursionDepth]
				.getInitialStates();

		for (State init : initialStates) {
			State newStateArray[] = new State[stateArray.length + 1];
			System
					.arraycopy(stateArray, 0, newStateArray, 0,
							stateArray.length);

			newStateArray[recursionDepth] = init;

			createInitialStateTuples_aux(recursionDepth + 1, newStateArray,
					fsm, checker, initStates);
		}

	}

	// returns a recursively generated Collection containing all possible
	// combinations of successors with a given symbol of the synchronization
	// alphabet
	private static Collection<StateTuple> getAllSuccessorsSync(
			String syncSymbol, StateTuple currentTuple,
			Collection<String> alphabets[]) {

		// If one of the fsms or the checker has the symbol in its alphabet but
		// its current state does not transition
		// then the Collection of successorstates is empty
		if ((currentTuple.checker != null)
				&& currentTuple.checker.transition(syncSymbol) == null)
			return new ArrayList<StateTuple>(0);

		for (int i = 0; i < currentTuple.states.length; i++) {
			if (alphabets[i].contains(syncSymbol)
					&& (currentTuple.states[i].transition(syncSymbol) == null))
				return new ArrayList<StateTuple>(0);

		}

		ArrayList<StateTuple> result = new ArrayList<StateTuple>();

		getAllSuccessorsSync_aux(0, new State[0], syncSymbol, currentTuple,
				alphabets, result);

		return result;
	}

	private static void getAllSuccessorsSync_aux(int recDepth, State[] states,
			String syncSymbol, StateTuple currentTuple,
			Collection<String> alphabets[], Collection<StateTuple> resultList) {

		if (recDepth >= currentTuple.states.length) {
			if (currentTuple.checker != null) {
				// we know that the checker must transition, because we made
				// sure of that
				// in the getAllSuccessorsSync method
				for (State checkerSucc : currentTuple.checker
						.transition(syncSymbol)) {
					resultList.add(new StateTuple(states, checkerSucc));
				}

			} else {
				resultList.add(new StateTuple(states, null));
			}
			return;
		}

		if (alphabets[recDepth].contains(syncSymbol)) {
			// if the fsm[recDepth] has the symbol in its alphabet
			// then we do all possible transitions and invoke the method
			// recursively.
			// We made sure in the getAllSuccessorsSync method, that it actually
			// does transition.
			for (State s : currentTuple.states[recDepth].transition(syncSymbol)) {
				State[] newStates = new State[recDepth + 1];
				System.arraycopy(states, 0, newStates, 0, recDepth);
				newStates[recDepth] = s;

				getAllSuccessorsSync_aux(recDepth + 1, newStates, syncSymbol,
						currentTuple, alphabets, resultList);
			}
		} else {
			// If the current fsm has the sync symbol not in its alphabet it
			// just stays
			// at its current state
			State[] newStates = new State[recDepth + 1];
			System.arraycopy(states, 0, newStates, 0, recDepth);
			newStates[recDepth] = currentTuple.states[recDepth];

			getAllSuccessorsSync_aux(recDepth + 1, newStates, syncSymbol,
					currentTuple, alphabets, resultList);
		}

	}

	/**
	 * Returns the output mode
	 * 
	 * @return OUTPUT_OFF, OUTPUT_NORMAL or OUTPUT_VERBOSE
	 */
	public static int getOutputMode() {
		return outputMode;
	}

	/**
	 * The partial order reduction constructs a reduced state graph of
	 * the parallel composition of all FSMs given in the array "fsm". This
	 * reduced state graph should behave equivalently to the full state
	 * graph when checking the conformity to a specification using a checker-
	 *  
	 * If a "checker" != null is specified, the synchronous product 
	 * (see product()) of the reduced state graph and the checker is
	 * constructed on the fly. Else ("checker" == null) an invariant 
	 * checker is assumed, and only the reduced state graph is constructed.
	 *    
	 * Note that if the checker is not invariant to the local symbols
	 * of the parallel components, the reduced state graph increases in size. 
	 * If the function is called with "checker"==null an invariant checker is 
	 * assumed, therefore doing a synchronous product with the reduced state
	 * graph and a non-invariant checker afterwards can easily lead to wrong results.
	 * 
	 * 
	 * An explanation of the algorithm is given in the paper "FSMCalc" 
	 * accompanying this program.  
	 * 
	 * @param checker
	 *            The fsm that's acting as the checker. May be null if no
	 *            checker is needed.
	 * @param fsm
	 *            The parallel components
	 * @return 
	 * 			  The reduced state graph of the product of the checker-fsm with the
	 * 			  parallelization of the other fsms.
	 * 
	 */

	@SuppressWarnings("unchecked")
	public static FiniteStateMachine partialOrderReduction(
			FiniteStateMachine checker, FiniteStateMachine... fsm) {
		if ((fsm == null))
			throw new IllegalArgumentException("Argument fsm may not be null");

		for (FiniteStateMachine f : fsm) {
			if (f == null)
				throw new IllegalArgumentException(
						"Argument fsm may not contain null");
		}

		if (fsm.length == 0)
			throw new IllegalArgumentException(
					"At least one fsm must be specified");
		
		if (outputMode != OUTPUT_OFF)
			System.err
					.println("starting partial order reduction with "
							+ fsm.length
							+ " parallel fsms and "
							+ (checker == null ? "without a checker"
									: "a checker"));

		FiniteStateMachine result = new FiniteStateMachine();

		/*
		 * build the synchronization alphabet (symbols that occur in more than
		 * one fsm) and the global alphabet (all symbols that appear in any fsm)
		 */

		Set<String> alphabets[] = new Set[fsm.length];

		Set<String> syncAlphabet = new HashSet<String>();

		Set<String> globalAlphabet = new HashSet<String>();

		for (int i = 0; i < fsm.length; i++) {
			alphabets[i] = fsm[i].determineAlphabet();

			Set<String> intersection = new HashSet<String>(alphabets[i]);
			intersection.retainAll(globalAlphabet);

			globalAlphabet.addAll(alphabets[i]);

			// add to the synchronisation alphabet all symbols that are in the
			// current alphabet, but were already put into the global alphabet
			syncAlphabet.addAll(intersection);

		}

		// find out which local transitions are not invariant in the checker
		// noninvariant transitions are those local (asynchronous) transitions,
		// that
		// cause a state change in the checker. These must be treated as global
		// transitions

		if (checker != null) {
			int missingSymbols = 0;
			int localTransitions = 0;

			for (State s : checker.getStates()) {
				for (String symbol : globalAlphabet) {
					// if the transition is synchronous, then it will be
					// expanded fully anyway
					// no need to add to nonInvariantTransitions
					if (syncAlphabet.contains(symbol))
						continue;

					ArrayList<State> successors = s.transition(symbol);

					if (successors == null) {
						// transition is non-invariant if it is missing
						syncAlphabet.add(symbol);
						missingSymbols++;
					} else
						for (State succ : successors) {
							if (succ != s) {
								localTransitions++;
								// transition is non-invariant if it causes the
								// checker to change state
								syncAlphabet.add(symbol);
								break;
							}
						}
				}
			}

			if ((outputMode != OUTPUT_OFF)
					&& ((missingSymbols > 0) || (localTransitions > 0))) {
				System.err
						.println("WARNING: checker is not invariant to some local transitions, some local symbols have to be treated as global symbols");
				if (outputMode == OUTPUT_VERBOSE) {
					System.out
							.println("  "
									+ missingSymbols
									+ " transition symbols are missing in parts of the checker\n  "
									+ localTransitions
									+ " local transition symbols cause the checker to change state");
				}

			}
		}

		// 2) find out for each fsm (except checker) which states are local.
		Set<State> nonLocalStates[] = new HashSet[fsm.length];

		for (int i = 0; i < fsm.length; i++) {
			nonLocalStates[i] = new HashSet<State>();

			for (State s : fsm[i].getStates()) {
				for (String symbol : s.getTransitions().keySet()) {

					// if the state contains a transition with a non-local
					// symbol, add the state to the non-local states
					if (syncAlphabet.contains(symbol))
						nonLocalStates[i].add(s);
				}
			}
		}

		if (outputMode == OUTPUT_VERBOSE) {
			System.err.println("synchronisation alphabet: "
					+ syncAlphabet.toString());
			System.err.println("global alphabet: " + globalAlphabet.toString());
		}

		// The DFS expansion stack: before the successors of a state are pushed
		// on this stack
		// null is pushed. Later when popping the values, a null signifies, that
		// all successors
		// have been expanded.
		// This information is used in modelling the DFS-search stack, which
		// must be explicetely
		// constructed, due to the non-recursive implementation.
		Stack<StateTuple> expansionStack = new Stack<StateTuple>();

		// The DFS-search stack, containing additional info on whether a state
		// has been fully or partially expanded
		// This information is needed in order to determine, whether a state
		// which would not normally qualify for full expansion
		// must be fully expanded because of a non-expanded cycle.
		Stack<StateTupleExpandInfo> searchStack = new Stack<StateTupleExpandInfo>();

		// the mapping from statePairs to states
		HashMap<StateTuple, State> tuplesToStates = new HashMap<StateTuple, State>();

		expansionStack.addAll(createInitialStateTuples(fsm, checker));

		// up to now only inital statetuples are on the expansionstack
		// corresponding states must be created

		// DEBUG:
		if (outputMode == OUTPUT_VERBOSE)
			System.err.println("initial states:" + expansionStack.toString());

		for (StateTuple t : expansionStack) {
			// create state in resulting fsm

			boolean allFinal = true;
			for (State s : t.states) {
				if (!s.isFinal()) {
					allFinal = false;
					break;
				}
			}

			if (checker != null)
				allFinal = allFinal && t.checker.isFinal();

			StateType type;
			if (allFinal)
				type = StateType.INITIAL_FINAL_STATE;
			else
				type = StateType.INITIAL_STATE;

			State resultState = result.createState(t.toString(), type);

			tuplesToStates.put(t, resultState);
		}

		int partialExpansions = 0;
		int fullExpansions = 0;
		int fullExpDueToCycle = 0;

		if (outputMode == OUTPUT_VERBOSE) {
			System.err.println("starting main-loop");
		}
		// main loop
		mainloop: while (!expansionStack.empty()) {
			StateTuple currentTuple = expansionStack.pop();
			State currentState = tuplesToStates.get(currentTuple);

			/*
			 * At first we check if we have to change the seach stack. We always
			 * push a "null" before expanding a state, so if we encounter "null"
			 * here, it means we have successfully expanded all successors of
			 * the state topmost on the search-stack, and can therefore remove
			 * it.
			 */

			if (currentTuple == null) {

				searchStack.pop();

				continue;
			}

			/*
			 * if this null is popped again later, we will know that we can take
			 * this state of the search stack.
			 */
			expansionStack.push(null);

			/*
			 * Now is our chance to find a possibility for partial expansion. If
			 * one of the fsm's contains only local transitions in the currently
			 * observed state (and at least one of those), we can try a partial
			 * expansion.
			 */
			if (outputMode == OUTPUT_VERBOSE) {
				System.err
						.println("current tuple = " + currentTuple.toString());
			}

			for (int i = 0; i < fsm.length; i++) {

				if (!currentTuple.states[i].transition.isEmpty() 
					&& (!nonLocalStates[i].contains(currentTuple.states[i]))) {
					
					// put on searchstack
					searchStack.push(new StateTupleExpandInfo(
							currentTuple.states, currentTuple.checker, false));
					
					/*
					 * Here we have found a candidate for partial expansion.
					 * During this expansion it may turn out that our candidate
					 * actually needs a full expansion. This can be the case if
					 * the partial expansion would close a loop in the resulting
					 * fsm, where all states have only been partially expanded.
					 * 
					 * In order to check this the states on the searchStack have
					 * a flag "fullyExpanded" set. When the partial expansion
					 * links back to an already discovered state, we go down
					 * (from the top) through the stack and check if we find
					 * this state before we find a fully expanded state.
					 */

					if (outputMode == OUTPUT_VERBOSE) {
						System.err
								.println("  partially expanding state with tuple-index "
										+ i);
					}
					boolean needsFullExpansion = partialOrderReductionExpandOnlyLocal_aux(
							i, currentTuple, currentState, fsm, tuplesToStates,
							result, expansionStack, searchStack, syncAlphabet);

					if (needsFullExpansion) {
						if (outputMode == OUTPUT_VERBOSE) {
							System.err
									.println("  partial expansion closed a loop that had never been fully expanded");
							System.err.println("  expanding fully");
						}
						// we need a full expansion here, because apparently our
						// partial expansion has closed a cycle with only
						// partial expansions
						// what we do is the following:

						for (int p = 0; p < fsm.length; p++) {
							// we expand all other states in the currentTuple
							// partially
							if (i != p)
								partialOrderReductionExpandOnlyLocal_aux(p,
										currentTuple, currentState, fsm,
										tuplesToStates, result, expansionStack,
										searchStack, syncAlphabet);
						}

						// and we expand only the global / synchronized
						// transitions
						partialOrderReductionExpandOnlyGlobal_aux(currentTuple,
								currentState, fsm, tuplesToStates, result,
								expansionStack, syncAlphabet, alphabets);

						// now we have expanded all local transitions, and all
						// global transitions
						// ==> thus, we have a full expansion

						// what we need to do now is set the "fullyExpanded" -
						// flag for this expansion-state on our search stack
						// because we already put the current tuple on the
						// searchStack we need to pop it, change the value and
						// push it again

						StateTupleExpandInfo updatedTuple = searchStack.pop();
						updatedTuple.fullyExpanded = true;
						searchStack.push(updatedTuple);

						fullExpansions++;
						fullExpDueToCycle++;
					} else
						partialExpansions++;

					// We have successfully expanded a state and can start the
					// mainloop again
					continue mainloop;
				}

			}

			/*
			 * At this point we know that we can't expand locally. So a full
			 * expansion is in order.
			 */
			searchStack.push(new StateTupleExpandInfo(currentTuple.states,
					currentTuple.checker, true));
			fullExpansions++;

			if (outputMode == OUTPUT_VERBOSE) {
				System.err.println("  expanding globally");
			}

			partialOrderReductionExpandLocalAndGlobal_aux(currentTuple,
					currentState, fsm, tuplesToStates, result, expansionStack,
					searchStack, syncAlphabet, alphabets);

		}

		if (outputMode == OUTPUT_VERBOSE) {
			System.err.println("partial expansions: " + partialExpansions);
			System.err.println("full expansions:    " + fullExpansions
			+ " (" + fullExpDueToCycle
			+ " due to closing a never fully expanded cycle");
		}

		return result;
	}

	/*
	 * Expand all transitions
	 */
	private static void partialOrderReductionExpandLocalAndGlobal_aux(
			StateTuple currentTuple, State currentState,
			FiniteStateMachine fsm[], Map<StateTuple, State> tuplesToStates,
			FiniteStateMachine result, Stack<StateTuple> expansionStack,
			Stack<StateTupleExpandInfo> searchStack,
			Collection<String> syncAlphabet, Collection<String> alphabets[]) {

		// expand all locally
		for (int i = 0; i < fsm.length; i++)
			partialOrderReductionExpandOnlyLocal_aux(i, currentTuple,
					currentState, fsm, tuplesToStates, result, expansionStack,
					searchStack, syncAlphabet);

		// expand globally
		partialOrderReductionExpandOnlyGlobal_aux(currentTuple, currentState,
				fsm, tuplesToStates, result, expansionStack, syncAlphabet,
				alphabets);
	}

	/*
	 * Expand only the global / synchronous transitions
	 */
	private static void partialOrderReductionExpandOnlyGlobal_aux(
			StateTuple currentTuple, State currentState,
			FiniteStateMachine fsm[], Map<StateTuple, State> tuplesToStates,
			FiniteStateMachine result, Stack<StateTuple> expansionStack,
			Collection<String> syncAlphabet, Collection<String> alphabets[]) {

		symbol_loop: for (String symbol : syncAlphabet) {

			Collection<StateTuple> tuples = getAllSuccessorsSync(symbol,
					currentTuple, alphabets);

			for (StateTuple succTuple : tuples) {
				// lookup if we already encountered this tuple
				State succState = tuplesToStates.get(succTuple);

				if (succState == null) {
					// the state doesn't exist yet, so we create it

					// first we find out whether it's a final state (= all the
					// states in the tuple are final)
					// The state cannot be initial or initial_final, because all
					// these states have already
					// been created during initialization
					StateType type = StateType.FINAL_STATE;

					for (int i = 0; i < succTuple.states.length; i++) {
						if (!succTuple.states[i].isFinal()) {
							type = StateType.NORMAL_STATE;
							break;
						}
						
					}

					if ((succTuple.checker != null)
							&& !succTuple.checker.isFinal())
						type = StateType.NORMAL_STATE;

					succState = result.createState(succTuple.toString(), type);

					// now we set a mapping from the tuple to the state
					tuplesToStates.put(succTuple, succState);

					// and put it on the expansionStack
					expansionStack.push(succTuple);
				}

				currentState.addTransition(symbol, succState);

			}

		}

	}

	/*
	 * Expand only the local transitions of the fsm with the specified index
	 */
	private static boolean partialOrderReductionExpandOnlyLocal_aux(
			int fsmindex, StateTuple currentTuple, State currentState,
			FiniteStateMachine fsm[], Map<StateTuple, State> tuplesToStates,
			FiniteStateMachine result, Stack<StateTuple> expansionStack,
			Stack<StateTupleExpandInfo> searchStack,
			Collection<String> syncAlphabet) {
		boolean needsFullExpansion = false;

		for (Entry<String, ArrayList<State>> entry : currentTuple.states[fsmindex]
				.getTransitions().entrySet()) {
			// we don't expand global symbols here
			if (syncAlphabet.contains(entry.getKey()))
				continue;

			for (State s : entry.getValue()) {
				State[] succStates = new State[fsm.length];
				System.arraycopy(currentTuple.states, 0, succStates, 0,
						fsm.length);

				succStates[fsmindex] = s;

				// because transitions that cause state changes in the checker
				// are treated
				// as global symbols, we know that the checker will stay in the
				// same state
				StateTuple succTuple = new StateTuple(succStates,
						currentTuple.checker);

				State succState = tuplesToStates.get(succTuple);

				if (succState == null) {
					// the state doesn't exist yet, so we create it

					// first we find out whether it's a final state (= all the
					// states in the tuple are final)
					// The state cannot be initial or initial_final, because all
					// these states have already
					// been created during initialization
					StateType type = StateType.FINAL_STATE;

					for (int i = 0; i < succTuple.states.length; i++) {
						if (!succTuple.states[i].isFinal()) {
							type = StateType.NORMAL_STATE;
							break;
						}
					}

					if ((succTuple.checker != null)
							&& !succTuple.checker.isFinal())
						type = StateType.NORMAL_STATE;

					succState = result.createState(succTuple.toString(), type);

					// now we set a mapping from the tuple to the state
					tuplesToStates.put(succTuple, succState);

					// and put it on the expansionStack
					expansionStack.push(succTuple);
				} else {

					/*
					 * Because the state we encountered here was already
					 * created, we have to check if linking back to it closes a
					 * loop that has never fully been expanded.
					 * 
					 * In order to do this we traverse the searchstack from the
					 * top down until 1) we find the state ==> a full expansions
					 * is necessary because a cycle was closed 2) we find a
					 * state that has been fully expanded ==> even if the state
					 * is on the search stack linking back to it closes no
					 * unexpanded loop, so we don't need to expand fully. 3) we
					 * reach the bottom of the stack ==> no full expansio
					 * necessary, because the state is not on the search-stack,
					 * so no loop is closed.
					 * 
					 */

					for (int i = searchStack.size() - 1; i >= 0; i--) {
						StateTupleExpandInfo current = searchStack.get(i);

						if (current.fullyExpanded)
							break;
						if (current.equals(succTuple)) {
							needsFullExpansion = true;
							break;
						}
					}

				}

				currentState.addTransition(entry.getKey(), succState);

			}

		}

		return needsFullExpansion;
	}

	/**
	 * Sets the output mode
	 * 
	 * @param outputMode
	 *            OUTPUT_OFF, OUTPUT_NORMAL or OUTPUT_VERBOSE
	 */
	public static void setOutputMode(int outputMode) {
		if ((outputMode < 0) || (outputMode > 2))
			throw new IllegalArgumentException("illegal mode");

		FiniteStateMachine.outputMode = outputMode;
	}

	/**
	 * Creates an empty finite state machine
	 */
	public FiniteStateMachine() {

	}

	/**
	 * Copy constructor. Creates an fsm that is a deep copy of the supplied fsm.
	 * 
	 * @param fsm
	 */
	public FiniteStateMachine(FiniteStateMachine fsm) {
		Map<State, State> origToCopy = new TreeMap<State, State>();

		// Copying states
		State copiedState;
		for (State s : fsm.getStates()) {
			copiedState = new State(s.name, this);

			origToCopy.put(s, copiedState);
			stateMap.put(copiedState.name, copiedState);

		}

		// setting initial and final states
		for (State s : fsm.getInitialStates()) {
			initialState.add(origToCopy.get(s));
		}
		for (State s : fsm.getFinalStates()) {
			finalState.add(origToCopy.get(s));
		}

		// setting transitions
		for (State s : fsm.state) {
			copiedState = origToCopy.get(s);
			for (Entry<String, ArrayList<State>> e : s.transition.entrySet()) {
				ArrayList<State> copiedTransitions = new ArrayList<State>();
				for (State succ : e.getValue()) {
					copiedTransitions.add(origToCopy.get(succ));
				}
				copiedState.transition.put(e.getKey(), copiedTransitions);

			}
		}
	}
	
	

	/**
	 * Checks whether the FSM accepts the given input.
	 * 
	 * @param s
	 *            the input in varargs or array form
	 * @return True if the given sequence of transformations ends at a final
	 *         state, false otherwise
	 */
	public boolean accepts(String... s) {
		
		if (s == null)
			throw new IllegalArgumentException("argument may not be null");
		
		
		
		Set<SearchNode> h1 = new TreeSet<SearchNode>();
		
		for(State initState : initialState) {
			h1.add(new SearchNode(initState, null));
		}
		
		Set<SearchNode> h2 = new TreeSet<SearchNode>();

		for (String inp : s) {
			for (SearchNode searchNode : h1) {
				ArrayList<State> successor = searchNode.state.transition(inp);
				if (successor != null) {
					for(State succState : successor) {
						h2.add(new SearchNode(succState, searchNode));
					}
				}
					
			}

			Set<SearchNode> temp = h1;
			h1 = h2;
			h2 = temp;
			h2.clear();

		}

		for (SearchNode searchNode : h1) {
			if (searchNode.state.isFinal()) {
				if(outputMode != OUTPUT_OFF) {
					
					System.err.println("trace:");
					if(outputMode == OUTPUT_VERBOSE) {
						System.err.println("(printing only one possible trace)");
					}
					Stack<State> trace = new Stack<State>();
					
					while(searchNode != null) {
						trace.push(searchNode.state);
						searchNode = searchNode.parent;
					}
					
					System.err.println("-> " + trace.pop().getName());
					
					int counter = 0;
					while(!trace.empty()) {
						System.err.println(s[counter++] + " -> " + trace.pop().getName());
					}
					System.err.println();
				}
				
				return true;
			}
				
		}

		if(outputMode == OUTPUT_VERBOSE) 
			System.err.println("No path to accepting state with given transitions");
		
		return false;
	}

	public void addTransition(String stateName, String symbol,
			String... successors) {

		State s = getState(stateName);

		if (s == null)
			throw new FSMException("No such state: " + stateName);

		s.addTransition(symbol, successors);
	}

	/**
	 * Makes a deep copy of the FSM
	 */
	public Object clone() {
		return new FiniteStateMachine(this);
	}

	/**
	 * Inverts all final states (all non-final states become final, all final
	 * states become non-final) If the given fsm is not deterministic or not
	 * complete, the complement of the power fsm will be returned.
	 * 
	 * @return A clone of the FiniteStateMachine with inverted final states
	 */
	@SuppressWarnings("unchecked")
	public FiniteStateMachine complementLanguage() {
		FiniteStateMachine fsm;
			
		boolean complete = isComplete();
		boolean deterministic = isDeterministic();

		if (complete && deterministic)
			fsm = new FiniteStateMachine(this);
		else {
			if(outputMode != OUTPUT_OFF) {
				if(!complete) {
					if(!deterministic) 
						System.err.println("original FSM is neither deterministic nor complete");
					else 
						System.err.println("original FSM is not complete");
				} else {
					System.err.println("original FSM is not deterministic");
				}
				
				System.err.println("constructing power-FSM");
			} 
			
			fsm = power();
		}
			

			
		Collection<State> temp = new ArrayList<State>(fsm.getStates());
		temp.removeAll(fsm.getFinalStates());
		
		fsm.getFinalStates().clear();
		
		fsm.getFinalStates().addAll(temp);
		
		return fsm;
	}

	/**
	 * Creates a normal state that is automatically named. Created names are of
	 * the form AnX where X is an integer.
	 * 
	 * @return A new anonymous state.
	 */
	public State createAnonymousState() {
		String name = "An" + anonCounter;
		while (stateMap.containsKey(name)) {
			name = "An" + (++anonCounter);
		}
		State s = createState(name, StateType.NORMAL_STATE);
		anonCounter++;
		return s;
	}

	/**
	 * Creates a new state of the specified type in the fsm. States are
	 * identified based on their name, no two states of the same name can be
	 * added.<br>
	 * If a state of that name already exists an exception is thrown. <br>
	 * State names may not start with a lower-case letter
	 * 
	 * @param name
	 *            The name of the state to add (may not start with lower-case
	 *            letter
	 * @param type
	 *            An enumeration giving the type of the State, either
	 *            NORMAL_STATE,INITIAL_STATE,FINAL_STATE or INITIAL_FINAL_STATE
	 * @return The state that was created
	 */
	public State createState(String name, StateType type) {
		if (stateMap.containsKey(name))
			throw new FSMException("A state of that name already exists");
		if (Character.isLowerCase(name.charAt(0))) {
			throw new FSMException(
					"State names may not start with a lower-case letter");
		}
		State s = new State(name, this);
		stateMap.put(name, s);

		switch (type) {
		case NORMAL_STATE:
			break;
		case INITIAL_STATE:
			initialState.add(s);
			break;
		case FINAL_STATE:
			finalState.add(s);
			break;
		case INITIAL_FINAL_STATE:
			initialState.add(s);
			finalState.add(s);
			break;
		}
		return s;
	}

	/**
	 * Returns a set of all transition symbols of the FSM. Changes of the FSM
	 * are not reflected in the returned set. The FSM has to be traversed every
	 * time determineAlphabet() is called.
	 * 
	 * @return The alphabet of the FSM
	 */
	public Set<String> determineAlphabet() {
		TreeSet<String> alphabet = new TreeSet<String>();

		for (State s : state) {
			alphabet.addAll(s.transition.keySet());
		}

		return alphabet;

	}

	public Collection<State> getFinalStates() {
		return finalState;
	}

	public Collection<State> getInitialStates() {
		return initialState;
	}

	public State getState(String stateName) {
		return stateMap.get(stateName);
	}

	public Collection<State> getStates() {
		return state;
	}

	public StateType getType(State s) {
		if (finalState.contains(s)) {
			if (initialState.contains(s)) {
				return StateType.INITIAL_FINAL_STATE;
			} else {
				return StateType.FINAL_STATE;
			}
		} else {
			if (initialState.contains(s)) {
				return StateType.INITIAL_STATE;
			} else {
				return StateType.NORMAL_STATE;
			}
		}
	}

	public boolean isComplete() {
		Collection<String> alphabet = determineAlphabet();
		for (State s : state) {
			if (!s.transition.keySet().containsAll(alphabet)) {
				return false;
			}
		}
		return true;
	}

	public boolean isDeterministic() {
		for (State s : state) {

			for (Entry<String, ArrayList<State>> e : s.transition.entrySet()) {
				if (e.getValue().size() > 1)
					return false;
			}

		}
		
		return true;
	}
	
	/**
	 * 
	 * @return true if a reachable final-state exists
	 */
	public boolean hasReachableFinalState() {
		Queue<SearchNode> q = new LinkedList<SearchNode>();
		HashSet<State> hashSet = new HashSet<State>();
		int counter = 0;
		
		for(State s : getInitialStates()) {
			q.add(new SearchNode(s, null));
			hashSet.add(s);
		}
		
		while(!q.isEmpty()) {
			counter++;
			SearchNode current = q.remove();
			if(current.state.isFinal()) {
				if(outputMode == OUTPUT_OFF)
					return true;
				
				if(outputMode == OUTPUT_VERBOSE) {
					System.err.println("final-state found after " + counter + " traversals");
				}
				
				Stack<SearchNode> stack = new Stack<SearchNode>();
				while(current != null) {
					stack.push(current);
					current = current.parent;
				}
				
				System.err.println("trace:");
				System.err.println("-> " + stack.pop().state.name);
								
				while(!stack.empty()) {
					current = stack.pop();
					System.err.println(current.symbol + " -> " + current.state.getName());
				}
				
				return true;
			}
			
			for(Entry<String,ArrayList<State>> entry : current.state.getTransitions().entrySet()) {
				for(State s : entry.getValue()) {
					if(!hashSet.contains(s)) {
						hashSet.add(s);
						q.add(new SearchNode(s, entry.getKey(), current));						
					}
				}
			}
			
		}
		
		if(outputMode == OUTPUT_VERBOSE) {
			System.err.println("no reachable final-state found after " + counter + " traversed states");
		}
		return false;
	}
	
	/**
	 * Creates an fsm where some symbols are replaced by other symbols.
	 * @param symbols 
	 * An array of symbols. Each even symbol symbols[i] (i=0,2,4,...) is replaced
	 * by its successor-symbol symbols[i+1]. The array must be even sized.
	 * @return A clone of the fsm where some symbols have been replaced by others.
	 */
	public FiniteStateMachine link(String... symbols) {
		if(symbols == null) 
			throw new IllegalArgumentException("argument may not be null");
		if( (symbols.length & 1) != 0) 
			throw new IllegalArgumentException("argument count must be even");
		for(String s : symbols) 
			if( s == null ) 
				throw new IllegalArgumentException("no argument may be null");
		
		FiniteStateMachine result = new FiniteStateMachine(this);
			
		for(State s : result.getStates()) {
			
			TreeMap<String, ArrayList<State>> newTransitions = new TreeMap<String, ArrayList<State>>();
			
			transitionLoop:				
			for(Entry<String, ArrayList<State>> entry : s.getTransitions().entrySet()) {
				for(int i = 0; i < symbols.length; i+=2) {
					if(entry.getKey().equals(symbols[i])) {
						newTransitions.put(symbols[i+1], entry.getValue());
						continue transitionLoop;
					}
				}
				
				newTransitions.put(entry.getKey(), entry.getValue());
			}
			
			s.transition = newTransitions;
		}
		
		return result;
	}

	/**
	 * Returns a clone of the fsm whose states are all final. Such an fsm is in
	 * effect a labelled transition system.
	 * 
	 * @return A clone of the fsm where all states are final
	 */
	public FiniteStateMachine makeAllStatesFinal() {
		FiniteStateMachine result = new FiniteStateMachine(this);
		result.finalState.addAll(state);
		return result;
	}

	/**
	 * Adds a non-final error state to the fsm. If one of the fsm's states does
	 * not transition with a symbol contained in the fsm's alphabet, a
	 * transition with the symbol to the error state is added.
	 * 
	 * @return A clone of the FiniteStateMachine which has been made complete
	 */
	public FiniteStateMachine makeComplete() {
		Collection<String> alphabet = determineAlphabet();
		FiniteStateMachine result = new FiniteStateMachine(this);
		State errorState = result.createState("Error", StateType.NORMAL_STATE);

		for (State s : result.getStates()) {
			for (String symbol : alphabet) {
				if (s.transition(symbol) == null) {
					s.addTransition(symbol, errorState);
				}
			}
		}

		return result;
	}

	/**
	 * Tries to make the "checker"-FSM invariant to the alphabet of all other fsms given as arguments in the array "fsm".
	 * This means that in every state <i>s</i> of the "checker"-fsm a reflexive transition <i>s --sym-->s</i>
	 * with the symbol <i>sym</i> is added, if s previously had no transition with <i>sym</i> and <i>sym</i>
	 * is an element of one of the FSMs given in the array "fsm".
	 *  
	 * This method can be used as preparation for the partial order reduction, as an invariant checker results in
	 * a better reduction of the state graph. Note that this operation changes the language of the finite state machine.
	 *  
	 * @param fsm The fsms containing the alphabets, the resulting machine should be invariant to. 
	 * @return A clone of the current fsm which is invariant to all transitions in the alphabets of the arguments, except for those
	 * non-invariant transitions which already existed.
	 */
	public FiniteStateMachine makeInvariantTo(FiniteStateMachine... fsm) {
				
		Set<String> globalAlphabet = new TreeSet<String>();
		for(FiniteStateMachine f : fsm) {
			if(f == null) throw new IllegalArgumentException("arguments may not be null");
			
			globalAlphabet.addAll(f.determineAlphabet());
		}
		
		FiniteStateMachine result = new FiniteStateMachine(this);
		
		for(State s : result.getStates()) {
			for(String symbol : globalAlphabet) {
				
				if(s.transition(symbol)== null) {
					s.addTransition(symbol, s);
				}
				
			}
		}
		
		return result;
	}
	
	/**
	 * Minimizes the FSM. Non-deterministic FSMs are first made deterministic by
	 * constructing the Power-FSM.
	 * 
	 * @return The minimized FSM
	 */
	public FiniteStateMachine minimize() {
		// TODO: Depends on result.state being sorted

		if (state.size() < 2)
			return new FiniteStateMachine(this);

		FiniteStateMachine result;

		if (outputMode == OUTPUT_VERBOSE)
			System.err.println("starting minimalization");

		// if not deterministic and complete=> make deterministic and complete
		boolean deterministic = isDeterministic();
		boolean complete = isComplete();
		if (deterministic&&complete) {
			result = new FiniteStateMachine(this);
		} else {
			if(outputMode != OUTPUT_OFF) {
				if(!complete) {
					if(!deterministic) 
						System.err.println("original FSM is neither deterministic nor complete");
					else 
						System.err.println("original FSM is not complete");
				} else {
					System.err.println("original FSM is not deterministic");
				}
				
				System.err.println("constructing power-FSM");
			}
			
			result = power();
		}

		HashSet<StatePair> oldRelation;
		HashSet<StatePair> newRelation;

		oldRelation = new HashSet<StatePair>(state.size() * state.size(), 0.97f);

		/*
		 * Initialization: build a set of statepairs (s1,s2) so that all
		 * (final,final) state pairs and all (nonfinal,nonfinal) state pairs are
		 * included
		 */

		for (State s1 : result.state) {
			for (State s2 : result.state) {
				if (s2.compareTo(s1) >= 0)
					break; // to avoid redundancy only include statepairs
				// (s1,s2) so that s1<s2
				if (s1.isFinal()) {
					if (s2.isFinal()) {
						oldRelation.add(new StatePair(s2, s1));
					}
				} else {
					if (!s2.isFinal()) {
						oldRelation.add(new StatePair(s2, s1));
					}
				}

			}
		}

		newRelation = new HashSet<StatePair>();
		boolean changed ;

		/*
		 * Main algorithm: Iteratively build the "is equivalent to"
		 * relationship In every step all statepairs (s1,s2) in oldRelation (the
		 * last iteration's result) are checked, they are put into newRelation
		 * only if for every transition of s1, there is a equivalent transition
		 * with the same symbol in s2, so that the both successors states
		 * (succ1, succ2) are part of oldRelation.
		 */

		do {
			changed = false;
			newRelation.clear();

			for (StatePair sp : oldRelation) {
				// keep the sp in the new relation?
				boolean keep = true;
				
				//first we checka if all transitions in sp.s1 have equivalent transitions in sp.s2
				for (Entry<String, ArrayList<State>> e : sp.s1.transition
						.entrySet()) {
					String symbol = e.getKey();
					ArrayList<State> successorList1 = e.getValue();
					ArrayList<State> successorList2 = sp.s2.transition(symbol);

					// if state 2 doesn't transition with a symbol, state 1
					// transitions with
					// is cannot be miminized to this state. No further
					// calculations needed
					if (successorList2 == null) {
						keep = false;
						break;
					}

					for (State succ1 : successorList1) {
						boolean foundEquivalentSuccessor = false;
						for (State succ2 : successorList2) {
							// statepairs are saved in oldrelation in the form
							// (s1,s2) s1<s2
							StatePair succSP = succ1.compareTo(succ2) < 0 ? new StatePair(
									succ1, succ2)
									: new StatePair(succ2, succ1);
							if ((succ1 == succ2)
									|| oldRelation.contains(succSP)) {
								foundEquivalentSuccessor = true;
								break;
							}
						}

						if (!foundEquivalentSuccessor) {
							keep = false;
							break;
						}
					}

					if (!keep)
						break;
					
					

				}
				

				if (keep) {
					newRelation.add(sp);
				} else {
					changed = true;
				}
			}

			// swap the two sets
			HashSet<StatePair> temp = newRelation;
			newRelation = oldRelation;
			oldRelation = temp;

		} while (changed);

		// minimize the fsm according to the relationship
		for (StatePair sp : oldRelation) {

			for (State s : result.state) {
				for (ArrayList<State> succList : s.transition.values()) {

					if (succList.remove(sp.s2) && (!succList.contains(sp.s1)))
						succList.add(sp.s1);
				}
			}

			result.stateMap.remove(sp.s2.name);
		
			result.finalState.remove(sp.s2);
			result.initialState.remove(sp.s2);

		}


		return result;
	}

	/**
	 * Constructs a FiniteStateMachine that is the asynchronous composition of
	 * the two FSMs.<br>
	 * If a transition has a symbol that exists in both FSM's alphabets it is
	 * treated as a synchronous transition, else it is treated as an
	 * asynchronous transition. Synchronous transitions can only "fire" if both
	 * original FSMs fire with it's symbol in the original states. If for
	 * example the letter 's' is a synchronous transition symbol, then the state
	 * A|B of the resulting FSM (A parallel to B) can only transition with 's'
	 * if both 'A' and 'B' have an equivalent transition with the same symbol.
	 * In this case A and B would both change into their successor states. <br>
	 * If on the other hand 's' was an asynchronous transition of the first FSM
	 * the state A|B could fire a transition with 's', if the original state 'A'
	 * had a corresponding transition. Only 'A' would change into it's successor
	 * state, while 'B' would remain unchanged.<br>
	 * If two states A (FSM1) and B (FSM2) are final in the original FSMs the
	 * resulting parallel state will be final.
	 * 
	 * @param fsm
	 * @return A fsm that represents the asynchronous composition of the two
	 *         FSMs
	 */
	public FiniteStateMachine parallel(FiniteStateMachine fsm) {
		if (fsm == null)
			throw new IllegalArgumentException("Argument may not be null");

		FiniteStateMachine result = new FiniteStateMachine();

		// build the synchronization alphabet
		Set<String> syncAlphabet = determineAlphabet();
		syncAlphabet.retainAll(fsm.determineAlphabet());

		// Maps two states to their parallel-state
		HashMap<StatePair, State> statesToParallel = new HashMap<StatePair, State>();

		LinkedList<StatePair> expansionQueue = new LinkedList<StatePair>();

		for (State s1 : getInitialStates()) {
			for (State s2 : fsm.getInitialStates()) {
				StatePair initialStatePair = new StatePair(s1, s2);
				expansionQueue.add(initialStatePair);

				// Initializing StringBuffer to correct length
				StringBuffer parallelStateName = new StringBuffer(s1.name
						.length()
						+ s2.name.length() + 3);

				parallelStateName.append('{').append(s1.name).append('|')
						.append(s2.name).append('}');

				StateType type;
				if (s1.isFinal() && s2.isFinal())
					type = StateType.INITIAL_FINAL_STATE;
				else
					type = StateType.INITIAL_STATE;

				State parallelState = result.createState(parallelStateName
						.toString(), type);
				statesToParallel.put(initialStatePair, parallelState);

			}
		}

		// start main loop
		while (!expansionQueue.isEmpty()) {
			StatePair currentPair = expansionQueue.removeFirst();
			State currentState = statesToParallel.get(currentPair);

			// for all transition in s1
			for (Entry<String, ArrayList<State>> entry : currentPair.s1.transition
					.entrySet()) {
				if (syncAlphabet.contains(entry.getKey())) {
					// is a synchronized transition => do rendevouz
					ArrayList<State> successors2 = currentPair.s2
							.transition(entry.getKey());
					if (successors2 != null) {
						for (State s1 : entry.getValue()) {
							for (State s2 : successors2) {
								StatePair succStatePair = new StatePair(s1, s2);
								State succState = statesToParallel
										.get(succStatePair);
								if (succState == null) {
									// state was not yet created, create it
									StateType type;
									if (succStatePair.s1.isFinal()
											&& succStatePair.s2.isFinal())
										type = StateType.FINAL_STATE;
									else
										type = StateType.NORMAL_STATE;

									// Initializing StringBuffer to correct
									// length
									StringBuffer stateName = new StringBuffer(3
											+ succStatePair.s1.getName()
													.length()
											+ succStatePair.s2.getName()
													.length());

									stateName.append('{').append(
											succStatePair.s1.getName()).append(
											'|').append(
											succStatePair.s2.getName()).append(
											'}');

									succState = result.createState(stateName
											.toString(), type);

									statesToParallel.put(succStatePair,
											succState);

									expansionQueue.addLast(succStatePair);
								}
								// set transition
								currentState.addTransition(entry.getKey(),
										succState);
							}
						}
					}

				} else {
					// is a asynchronous transition => do interleaving
					for (State s1succ : entry.getValue()) {
						// s1 transitions, s2 stays the same
						StatePair succStatePair = new StatePair(s1succ,
								currentPair.s2);
						State succState = statesToParallel.get(succStatePair);
						if (succState == null) {
							// state was not yet created, create it
							StateType type;
							if (succStatePair.s1.isFinal()
									&& succStatePair.s2.isFinal())
								type = StateType.FINAL_STATE;
							else
								type = StateType.NORMAL_STATE;

							// Initializing StringBuffer to correct length
							StringBuffer stateName = new StringBuffer(3
									+ succStatePair.s1.getName().length()
									+ succStatePair.s2.getName().length());

							stateName.append('{').append(
									succStatePair.s1.getName()).append('|')
									.append(succStatePair.s2.getName()).append(
											'}');

							succState = result.createState(
									stateName.toString(), type);

							statesToParallel.put(succStatePair, succState);

							expansionQueue.addLast(succStatePair);
						}
						// set transition
						currentState.addTransition(entry.getKey(), succState);

					}
				}
			}

			// for all transitions in s2
			for (Entry<String, ArrayList<State>> entry : currentPair.s2.transition
					.entrySet()) {
				if (syncAlphabet.contains(entry.getKey()))
					continue; // this has already been taken care of in the
				// preceding loop
				else {
					// is an asynchronous transition => do interleaving
					for (State s2succ : entry.getValue()) {
						StatePair succStatePair = new StatePair(currentPair.s1,
								s2succ);
						State succState = statesToParallel.get(succStatePair);
						if (succState == null) {
							// state was not yet created, create it
							StateType type;
							if (succStatePair.s1.isFinal()
									&& succStatePair.s2.isFinal())
								type = StateType.FINAL_STATE;
							else
								type = StateType.NORMAL_STATE;

							// Initializing StringBuffer to correct length
							StringBuffer stateName = new StringBuffer(3
									+ succStatePair.s1.getName().length()
									+ succStatePair.s2.getName().length());

							stateName.append('{').append(
									succStatePair.s1.getName()).append('|')
									.append(succStatePair.s2.getName()).append(
											'}');

							succState = result.createState(
									stateName.toString(), type);

							statesToParallel.put(succStatePair, succState);

							expansionQueue.addLast(succStatePair);
						}
						// set transition
						currentState.addTransition(entry.getKey(), succState);

					}
				}
			}

		}

		return result;
	}

	/**
	 * Returns the power-fsm of the finite state machine.
	 * 
	 * @return The deterministic and complete power-fsm of the finite state machine.
	 */
	@SuppressWarnings("unchecked")
	public FiniteStateMachine power() {
		// init result
		FiniteStateMachine result = new FiniteStateMachine();

		// determine alphabet
		Collection<String> alphabet = determineAlphabet();

		// provides Mapping
		Map<Set<State>, State> powerStateMap = new HashMap<Set<State>, State>();

		// bfs expansion queue
		Queue<Set<State>> expansionQueue = new LinkedList<Set<State>>();

		// construct initial state, set mapping and put on expansion queue
		StringBuffer stateName = new StringBuffer();
		stateName.append('{');
		
		StateType initialStateType = StateType.INITIAL_STATE;
		
		for (State s : initialState) {
			stateName.append(s.name).append(',');
			if(s.isFinal()) initialStateType = StateType.INITIAL_FINAL_STATE;
		}

		if (stateName.length() > 1)
			stateName.deleteCharAt(stateName.length() - 1);

		stateName.append('}');

				
		State powerInitialState = result.createState(stateName.toString(),
				initialStateType);

		powerStateMap.put(initialState, powerInitialState);
		expansionQueue.add(initialState);

		// main bfs loop
		while (!expansionQueue.isEmpty()) {
			// get next set to expand
			Set<State> current = expansionQueue.remove();

			State currentState = powerStateMap.get(current);

			for (String symbol : alphabet) {
				// create the successors of the current set
				Set<State> successorSet = new TreeSet<State>();

				for (State s : current) {
					List successor = s.transition(symbol);
					if (successor != null)
						successorSet.addAll(successor);
				}

				State successorState = powerStateMap.get(successorSet);

				if (successorState == null) {
					// State did not yet exist
					stateName = new StringBuffer();
					stateName.append('{');

					boolean isFinal = false;
					for (State s : successorSet) {
						stateName.append(s.name).append(',');

						// check whether one of the states in the state set is a
						// final state
						if (!isFinal && s.isFinal())
							isFinal = true;
					}

					if (stateName.length() > 1)
						stateName.deleteCharAt(stateName.length() - 1);
					stateName.append('}');

					// create the state
					StateType type;
					if (isFinal)
						type = StateType.FINAL_STATE;
					else
						type = StateType.NORMAL_STATE;

					successorState = result.createState(stateName.toString(),
							type);

					powerStateMap.put(successorSet, successorState);
					expansionQueue.add(successorSet);
				}

				currentState.addTransition(symbol, successorState);

			}

		}



		return result;
	}



	/**
	 * Creates a product machine which accepts as a language the intersection of
	 * the two languages of the given finite state machines. <br>
	 * The resulting product machine is not a complete product, it contains only
	 * reachable states.
	 * 
	 * @param fsm
	 * @return A FSM representing the product between the two input FSMs.
	 */

	@SuppressWarnings("unchecked")
	public FiniteStateMachine product(FiniteStateMachine fsm) {
		if (fsm == null)
			throw new IllegalArgumentException("argument may not be null");

		FiniteStateMachine result = new FiniteStateMachine();

		// Provides mapping from two statenames to their product state
		HashMap<StatePair, State> productStatesMap = new HashMap<StatePair, State>();

		// queue for bfs-like product algorithm
		Queue<StatePair> bfsqueue = new LinkedList<StatePair>();

		// add initial-states to queue
		for (State s1 : initialState) {
			for (State s2 : fsm.initialState) {

				// Initializing StringBuffer to correct length
				StringBuffer productStateName = new StringBuffer(s1.getName()
						.length()
						+ s2.getName().length() + 3);

				productStateName.append('{').append(s1.name).append(",")
						.append(s2.name).append("}");

				StatePair sp = new StatePair(s1, s2);

				StateType type;

				if (s1.isFinal() && s2.isFinal())
					type = StateType.INITIAL_FINAL_STATE;
				else
					type = StateType.INITIAL_STATE;

				State productState = result.createState(productStateName
						.toString(), type);

				// Add initial states to product state mapping (StatePair ==>
				// State)
				productStatesMap.put(sp, productState);

				// Add initial states to bfs queue for further expansion in main
				// algorithm
				bfsqueue.add(sp);
			}
		}

		/*
		 * Here follows the main part of the product algorithm Removes from
		 * queue and expands statepairs into their product successors, when both
		 * states of the pair transition with the same symbol. The succesors are
		 * put on the queue again.
		 */
		while (!bfsqueue.isEmpty()) {
			StatePair current = bfsqueue.remove();
			for (Entry<String, ArrayList<State>> e : current.s1.transition
					.entrySet()) {

				ArrayList<State> s2Transitions = current.s2.transition(e
						.getKey());

				if (s2Transitions == null)
					continue;

				for (State s1 : e.getValue()) {
					// for all states current.s1 transitions to with symbol
					// e.getKey()

					for (State s2 : s2Transitions) {
						// for all states current.s2 transitions to with the
						// same symbol

						// check whether productState of the two successor was
						// already expanded
						StatePair productsp = new StatePair(s1, s2);
						State productState = productStatesMap.get(productsp);

						if (productState == null) {
							// This successor product state of (s1/s2) was not
							// yet reached / expanded.
							// Create state, set mapping, put on queue for
							// expansion, add to result-fsm

							// Initializing StringBuffer to correct length
							StringBuffer productStateName = new StringBuffer(s1
									.getName().length()
									+ s2.getName().length() + 3);
							productStateName.append('{').append(s1.name)
									.append(",").append(s2.name).append("}");

							StateType type;
							if (s1.isFinal() && s2.isFinal())
								type = StateType.FINAL_STATE;
							else
								type = StateType.NORMAL_STATE;

							productState = result.createState(productStateName
									.toString(), type);

							productStatesMap.put(productsp, productState);

							bfsqueue.add(productsp);

						}
						// set the product transition
						State currentState = productStatesMap.get(current);
						currentState.addTransition(e.getKey(), productState);
					}
				}

			}
		}

		return result;
	}

	/**
	 * Removes a state (also initial state and final state) from the FSM. All
	 * connections to the state are also removed.
	 * 
	 * @param toRemove
	 *            The state to be removed. May not be null.
	 * @return True if s could be removed, false if no such state existed
	 */

	public boolean removeState(State toRemove) {
		if (toRemove == null)
			return false;

		for (State s : state) {
			s.removeTransitions(toRemove);
		}

		initialState.remove(toRemove);
		finalState.remove(toRemove);

		State removed = stateMap.remove(toRemove.name);
		return removed != null;

	}

	/**
	 * Sets the state to the specified type. If no such state exists an
	 * exception is thrown.
	 * 
	 * @param s
	 * @param type
	 */
	public void setType(State s, StateType type) {

		if (!stateMap.containsKey(s.name))
			throw new IllegalArgumentException("Specified state is not in fsm");

		switch (type) {
		case NORMAL_STATE:
			initialState.remove(s);
			finalState.remove(s);
			break;
		case INITIAL_STATE:
			finalState.remove(s);
			initialState.add(s);
			break;
		case FINAL_STATE:
			finalState.add(s);
			initialState.remove(s);
			break;
		case INITIAL_FINAL_STATE:
			initialState.add(s);
			finalState.add(s);
			break;
		}
	}

	@Override
	public String toString() {

		StringBuffer b = new StringBuffer();

		for (State s : state) {
			if (initialState.contains(s))
				b.append('i');
			if (finalState.contains(s))
				b.append('f');
			b.append(s.getStringDescription());
			b.append('\n');
		}
		return b.toString();
	}

	public void visualize() {
		new FSMViewer(this, "fsm");
	}
}

/* used for mapping pairs of states to a product-state / parallel state */
class StatePair {
	State s1;

	State s2;

	public StatePair(State s1, State s2) {
		this.s1 = s1;
		this.s2 = s2;

	}

	@Override
	public boolean equals(Object obj) {
		StatePair sp = (StatePair) obj;
		return sp.s1.equals(s1) && sp.s2.equals(s2);
	}

	@Override
	public int hashCode() {
		int hash1 = s1.name.hashCode();
		int hash2 = s2.name.hashCode();
		int tmp = 0xFFFF0000 & hash2;

		int result = (hash1 ^ (hash2 << 16)) ^ (tmp >>> 16);

		return result;
	}
}

class StateTuple {
	public State checker;

	public State states[];

	public StateTuple(State[] states, State checker) {
		this.states = states;
		this.checker = checker;
	}

	@Override
	public boolean equals(Object obj) {
		StateTuple compared = (StateTuple) obj;

		if (compared.states.length != states.length)
			return false;

		for (int i = 0; i < states.length; i++)
			if (!states[i].equals(compared.states[i]))
				return false;

		return (checker == null) || checker.equals(compared.checker);
	}

	@Override
	public int hashCode() {
		int hash = states[0].getName().hashCode();

		for (int i = 1; i < states.length; i++) {
			int tmp = (0xFFFF & hash) >>> 16;
			hash = ((hash << 16) ^ states[i].getName().hashCode()) ^ tmp;
		}

		if (checker != null) {
			int tmp = (0xFFFF & hash) >>> 16;
			hash = ((hash << 16) ^ checker.getName().hashCode()) ^ tmp;
		}

		return hash;
	}

	public String toString() {
		int length = 0;
		for (State s : states) {
			length += s.getName().length();
		}
		length += checker == null ? states.length + 1 : checker.getName()
				.length()
				+ states.length + 2; // Brackets and commas;

		StringBuffer nameBuffer = new StringBuffer(length);
		nameBuffer.append('{');

		if (checker != null)
			nameBuffer.append(checker.getName()).append(',');

		for (State s : states)
			nameBuffer.append(s.getName()).append('|');

		nameBuffer.deleteCharAt(length - 1);
		nameBuffer.append('}');

		return nameBuffer.toString();
	}
}

class StateTupleExpandInfo extends StateTuple {
	public boolean fullyExpanded;

	public StateTupleExpandInfo(State state[], State checker,
			boolean fullyExpanded) {
		super(state, checker);

		this.fullyExpanded = fullyExpanded;
	}
}

class SearchNode implements Comparable {
	public SearchNode parent;
	public State state;
	public String symbol;
	
	public SearchNode(State s, SearchNode parent) {
		state = s;
		this.parent = parent;
	}
	
	public SearchNode(State s, String symbol, SearchNode parent) {
		this(s, parent);
		this.symbol = symbol;	
	}

	public int compareTo(Object o) {
		return state.compareTo(((SearchNode) o).state);
	}
	
	@Override
	public boolean equals(Object obj) {
		return state.equals( ((SearchNode) obj).state);
	}	
	
	@Override
	public int hashCode() {
		return state.getName().hashCode();
	}
}