package core;

import java.util.ArrayList;
import java.util.Map;

import java.util.TreeMap;

import java.util.Map.Entry;

/**
 * This class represents a single state of a FSM, the components being its name
 * and its transitions.
 * 
 * @author Leopold Haller
 */

public class State implements Comparable {

	protected FiniteStateMachine fsm;

	protected String name;

	protected TreeMap<String, ArrayList<State>> transition;


	/**
	 * Initialzizes a state with a given name. Use
	 * FiniteStateMachine.createState instead.
	 * 
	 * @param name
	 *            The name of the state
	 * @param fsm
	 *            The fsm the state belongs to.
	 * @param anon
	 *            Specifies whether the State is anonymous.
	 */
	protected State(String name, FiniteStateMachine fsm) {
		this.fsm = fsm;
		this.name = name;
		transition = new TreeMap<String, ArrayList<State>>();
	}

	/**
	 * Adds a transition to the State, the symbol is automatically added to the
	 * state's finite state machine
	 * 
	 * @param symbol
	 *            The symbol with which the state should transition
	 * @param s
	 *            The successor states reachable through transitioning with
	 *            symbol
	 */
	public void addTransition(String symbol, State... s) {
		if (symbol == null)
			throw new IllegalArgumentException("symbol may not be null");
		if (s == null)
			throw new IllegalArgumentException("state may not be null");
		

		ArrayList<State> successors = transition.get(symbol);
		if (successors == null) {
			// Transition not yet defined
			successors = new ArrayList<State>(s.length);
			for (State ss : s)
				successors.add(ss);
			transition.put(symbol, successors);
		} else {
			// Transition already defined
			for (State ss : s)
				successors.add(ss);
		}

	}

	public void addTransition(String symbol, String... s) {
		if (symbol == null)
			throw new IllegalArgumentException("symbol may not be null");
		if (s == null)
			throw new IllegalArgumentException("state may not be null");
		if (s.length == 0)
			return;

		ArrayList<State> successors = transition.get(symbol);
		if (successors == null) {
			// Transition not yet defined
			successors = new ArrayList<State>(s.length);
			for (String ss : s) {
				State succState = fsm.getState(ss);
				
				if(succState == null)
					throw new FSMException("no such state; " + ss);
				
				successors.add(succState);
			}			
				
			transition.put(symbol, successors);
		} else {
			// Transition already defined
			for (String ss : s)
				successors.add(fsm.getState(ss));
		}

	}
	
	public Map<String, ArrayList<State>> getTransitions() {
		return transition;
	}
	
	public int compareTo(Object o) {
		State s;
		try {
			s = (State) o;
		} catch (ClassCastException e) {
			throw new IllegalArgumentException(
					"States can only be compared to other states");
		}
		return name.compareTo(s.name);
	}

	@Override
	public boolean equals(Object obj) {
		State s = (State) obj;
		return s.name.equals(name);
	}

	public String getName() {
		return name;
	}

	public boolean isFinal() {
		return fsm.getFinalStates().contains(this);
	}

	public boolean isInitial() {
		return fsm.getInitialStates().contains(this);
	}

	public boolean removeTransition(String symbol) {
		if (symbol == null)
			throw new IllegalArgumentException("symbol may not be null");
		ArrayList<State> succ = transition.remove(symbol);
		return succ != null;
	}

	public boolean removeTransition(String symbol, State s) {
		if (symbol == null)
			throw new IllegalArgumentException("symbol may not be null");
		if (s == null)
			throw new IllegalArgumentException("state may not be null");
		ArrayList<State> toRemove;
		toRemove = transition.get(symbol);

		if (toRemove == null)
			return false;
		if (toRemove.size() <= 1) {
			transition.remove(symbol);
		} else {
			toRemove.remove(s);
		}

		return true;
	}

	/**
	 * Removes all transitions to the given state
	 * 
	 * @param toRemove
	 *            The state to which all transitions are to be removed
	 * @return True if the state (this) had transitions to the state s.
	 */
	public boolean removeTransitions(State toRemove) {
		boolean hadTransitions = false;
		ArrayList<String> removeableTransitions = new ArrayList<String>();
		for (Entry<String, ArrayList<State>> e : transition.entrySet()) {
			hadTransitions = hadTransitions || (e.getValue().remove(toRemove));
			if (e.getValue().size() == 0)
				removeableTransitions.add(e.getKey());
		}

		// remove empty transition associations (where the size of the successor
		// arraylist==0)
		for (String s : removeableTransitions) {
			transition.remove(s);
		}

		return hadTransitions;
	}

	public void setFinal(boolean finalState) {
		if (finalState) {
			if(isInitial())
				fsm.setType(this, StateType.INITIAL_FINAL_STATE);
			else
				fsm.setType(this, StateType.FINAL_STATE);
		} else {
			if(isInitial())
				fsm.setType(this, StateType.INITIAL_STATE);
			else
				fsm.setType(this, StateType.NORMAL_STATE);
		}
	}

	public void setInitial(boolean initialState) {
		if (initialState) {
			if(isFinal())
				fsm.setType(this, StateType.INITIAL_FINAL_STATE);
			else
				fsm.setType(this, StateType.INITIAL_STATE);
		} else {
			if(isFinal())
				fsm.setType(this, StateType.FINAL_STATE);
			else
				fsm.setType(this, StateType.NORMAL_STATE);
		}
	}

	public String getStringDescription() {
		StringBuffer b = new StringBuffer();

		// if it's anonymous return the string of transition-symbols until next
		// non-anonymous state
		/*if (anonymous) {
			
			for (Entry<String, ArrayList<State>> entry : transition.entrySet()) {
				b.append(entry.getKey());
				b.append('.');
				assert entry.getValue().size() == 1;
				State succ = entry.getValue().get(0);
				if (succ.isAnonymous())
					b.append(succ.getStringDescription());
				else
					b.append(succ.getName());
			}
			
			return b.toString();
		}*/

		// if it's not anonymous, return the State in the form Name =
		// symbol.Successor + symbol2.Successor2 + ...
		b.append(name).append(" =");
		for (Entry<String, ArrayList<State>> e : transition.entrySet()) {
			for (State s : e.getValue()) {
				b.append(' ');
				b.append(e.getKey());
				b.append('.');

				/*if (s.isAnonymous()) {
					b.append(s.getStringDescription());
				} else {*/
					b.append(s.name);
				//}
				b.append(" +");

			}
		}

		b.deleteCharAt(b.length() - 1);
		return b.toString();
	}

	public String toString() {
		return getName();
	}

	/**
	 * Returns the successors of the state with the given symbol, null if none
	 * exist.
	 * 
	 * @param symbol
	 *            The symbol to transition with.
	 * @return An arraylist containing the successors, changes in the list
	 *         reflect in the transitions
	 */
	public ArrayList<State> transition(String symbol) {
		return transition.get(symbol);
	}
	
	
}
