package gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.Map.Entry;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.UIManager;

import org.jgraph.JGraph;
import org.jgraph.graph.DefaultCellViewFactory;
import org.jgraph.graph.DefaultEdge;
import org.jgraph.graph.DefaultGraphCell;
import org.jgraph.graph.DefaultGraphModel;
import org.jgraph.graph.DefaultPort;
import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.GraphLayoutCache;
import org.jgraph.graph.GraphModel;

import core.FiniteStateMachine;
import core.State;

/**
 * Simple implementation of a visualizer
 * @author Leopold Haller
 */
public class FSMViewer implements ActionListener {
	private static final Color FINAL_STATE_COLOR = new Color(0xF05555);
	private static final Color INITIAL_STATE_COLOR = new Color(0x5555F0);
	private static final Color INITIAL_FINAL_STATE_COLOR = new Color(0xF055F0);
	private static final Color NORMAL_STATE_COLOR = Color.ORANGE;
	private static final int WINDOW_WIDTH = 640;
	private static final int WINDOW_HEIGHT = 480;
	private static final int X_START = 50;
	private static final int Y_START = 50;
	private static final int LINE_SPACING = 100;
	private static final int COL_SPACING = 100;
	
	
	private JFrame frame;
	private GraphModel model;
	private GraphLayoutCache view;
	private JGraph graph;
	private ArrayList<DefaultGraphCell> cells;
	private HashMap<State, DefaultGraphCell> stateToCell;
	private FiniteStateMachine fsm;
	private ArrayList<DefaultEdge> edges;
	private ArrayList<DefaultEdge> selfLoops;
	private HashMap<StatePair, DefaultEdge> edgeMap;
	private HashMap<State, DefaultPort> portMap;
	private JMenuBar menuBar;	
	private JMenu optionsMenu;
	private JMenu fileMenu;
	private JMenuItem itemLinesStraight;
	private JMenuItem itemLinesSpline;
	private JMenuItem itemRoutingOff;
	private JMenuItem itemRoutingOn;
	private JMenuItem itemFileExit;
		
	public FSMViewer(FiniteStateMachine fsm1, String name) {
		try {
	        UIManager.setLookAndFeel(
	            UIManager.getCrossPlatformLookAndFeelClassName());
	    } catch (Exception e) { }
		
	    this.fsm = fsm1;
	    
		frame = new JFrame("FSM Viewer - " + name );
		
		
		setupMenu();
		
		
		model = new DefaultGraphModel();
		
		view = new GraphLayoutCache(
									model, 
									new DefaultCellViewFactory());
		graph = new JGraph(model, view);
		
		graph.setEditable(false);
		graph.setAntiAliased(true);
		graph.setDisconnectable(false);
		
		
		stateToCell = new HashMap<State, DefaultGraphCell>((int) (1.5f*fsm.getStates().size()), 2f);
						
		cells = new ArrayList<DefaultGraphCell>(fsm.getStates().size());
		
		
		for(State s : fsm.getStates()) {
			DefaultGraphCell cell = new DefaultGraphCell(s);
			cells.add(cell);
						
			stateToCell.put(s, cell);
			
			Map attributes = cell.getAttributes();
			
			Color c = null;
			
			switch(fsm.getType(s)) {
			case INITIAL_FINAL_STATE: c = INITIAL_FINAL_STATE_COLOR; break;
			case INITIAL_STATE: c = INITIAL_STATE_COLOR; break;
			case FINAL_STATE: c = FINAL_STATE_COLOR; break;
			case NORMAL_STATE: c = NORMAL_STATE_COLOR; break;
				
			}
			GraphConstants.setGradientColor(attributes, c);
			
			
			GraphConstants.setOpaque(attributes, true);
			GraphConstants.setBorder(attributes, BorderFactory.createLineBorder(Color.BLACK, 2));
			GraphConstants.setDisconnectable(attributes, false);
			GraphConstants.setAutoSize(attributes, true);
			
			GraphConstants.setBounds(attributes, new Rectangle2D.Double((int)(Math.random()*WINDOW_WIDTH),(int)(Math.random()*WINDOW_HEIGHT),50,50 ) );
		}
		
		
		layoutBFS();
		
		//Setup connections (edges)
		edges = new ArrayList<DefaultEdge>();
		selfLoops = new ArrayList<DefaultEdge>();
		edgeMap = new HashMap<StatePair, DefaultEdge>();
		portMap = new HashMap<State, DefaultPort>();
		
		for(State s : fsm.getStates()) {
			for(Entry<String, ArrayList<State>> entry : s.getTransitions().entrySet()) {
				for(State succ : entry.getValue()) {
					DefaultGraphCell sourceCell = stateToCell.get(s);
					DefaultGraphCell destCell = stateToCell.get(succ);
					
					DefaultPort inPort = portMap.get(succ);
					if(inPort == null) {
						inPort = new DefaultPort();
						portMap.put(succ, inPort);
						destCell.add(inPort);
						inPort.setParent(destCell);
					}
					
					DefaultPort outPort = portMap.get(s);
					if(outPort == null) {
						outPort = new DefaultPort();
						portMap.put(s, outPort);
						sourceCell.add(outPort);
						outPort.setParent(sourceCell);
					}
					
					sourceCell.add(outPort);
					destCell.add(inPort);
					
					outPort.setParent(sourceCell);
					inPort.setParent(destCell);
					
					DefaultEdge edge = edgeMap.get(new StatePair(s, succ));
					if(edge == null) {
						edge = new DefaultEdge(entry.getKey());
						edgeMap.put(new StatePair(s,succ), edge);
					} else { 
						edge.setUserObject( ((String) edge.getUserObject()) + ", " + entry.getKey() );
					}
					
					edge.setSource(outPort);
					edge.setTarget(inPort);
					
					GraphConstants.setLineEnd(edge.getAttributes(), GraphConstants.ARROW_CLASSIC);
					GraphConstants.setEndFill(edge.getAttributes(), true);
					
					if(s == succ) {
						GraphConstants.setLineStyle(edge.getAttributes(), GraphConstants.STYLE_SPLINE);
						GraphConstants.setRouting(edge.getAttributes(), GraphConstants.ROUTING_SIMPLE);
						selfLoops.add(edge);
					}
					GraphConstants.setLabelPosition(edge.getAttributes(), new Point2D.Float(GraphConstants.PERMILLE/6,-5));
					edges.add(edge);
				}
			}
		}
		
		graph.getGraphLayoutCache().insert(cells.toArray());
		graph.getGraphLayoutCache().insert(edges.toArray());

		frame.getContentPane().add(new JScrollPane(graph));
		
		//frame.pack();
		frame.setSize(new Dimension(WINDOW_WIDTH, WINDOW_HEIGHT));
		frame.setVisible(true);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	}

	private void setupMenu() {
		menuBar = new JMenuBar();
		
		fileMenu = new JMenu("file");
		fileMenu.setMnemonic('f');
		itemFileExit = new JMenuItem("exit");
		itemFileExit.setMnemonic('x');
		
		itemFileExit.addActionListener(this);
		
		fileMenu.add(itemFileExit);
		
		menuBar.add(fileMenu);
		
		optionsMenu = new JMenu("options");
		optionsMenu.setMnemonic('o');
		
		ButtonGroup lineGroup = new ButtonGroup();
				
		itemLinesStraight = new JRadioButtonMenuItem("straight edges", true);
		itemLinesStraight.setMnemonic('s');
		itemLinesSpline = new JRadioButtonMenuItem("curved edges");
		itemLinesSpline.setMnemonic('c');
						
		lineGroup.add(itemLinesStraight);
		lineGroup.add(itemLinesSpline);
		
		ButtonGroup routingGroup = new ButtonGroup();
		
		itemRoutingOff = new JRadioButtonMenuItem("no edge routing", true);
		itemRoutingOff.setMnemonic('n');
		itemRoutingOn = new JRadioButtonMenuItem("routing on");
		itemRoutingOn.setMnemonic('r');
		
		routingGroup.add(itemRoutingOff);
		routingGroup.add(itemRoutingOn);
		
			
		optionsMenu.add(itemRoutingOff);
		optionsMenu.add(itemRoutingOn);
		
		optionsMenu.add(new JSeparator());
		
		optionsMenu.add(itemLinesStraight);
		optionsMenu.add(itemLinesSpline);
		
		itemLinesStraight.setEnabled(false);
		itemLinesSpline.setEnabled(false);
		
		
		menuBar.add(optionsMenu);
		
		
		
		
		frame.setJMenuBar(menuBar);
		
		
		//add this as listener of menuitems
		
		itemLinesStraight.addActionListener(this);
		itemLinesSpline.addActionListener(this);
		itemRoutingOff.addActionListener(this);
		itemRoutingOn.addActionListener(this);
		
	}

	private void layoutBFS() {
		//start BFS Layout
		Set<State> reached = new HashSet<State>(fsm.getStates().size());
		Queue<BFSNode<State>> expansionQueue = new LinkedList<BFSNode<State>>();
						
		for(State s : fsm.getInitialStates()) {
			reached.add(s);
			expansionQueue.add(new BFSNode<State>(s, 0));			
		}
		
		
		int xpos;
		int ypos;
		xpos = X_START;
		ypos = Y_START;
		int lastDepth = 0;
		
		while(!expansionQueue.isEmpty()) {
			BFSNode<State> currentNode = expansionQueue.remove();
			if(currentNode.depth > lastDepth) {
				lastDepth = currentNode.depth;
				xpos = X_START;
				ypos += LINE_SPACING;
			}
			for(ArrayList<State> successors : currentNode.value.getTransitions().values())
				for(State succ : successors) {
					if(reached.contains(succ))
						continue;
					reached.add(succ);
					expansionQueue.add(new BFSNode<State>(succ, lastDepth + 1));
				}
			
			GraphConstants.setBounds(stateToCell.get(currentNode.value).getAttributes(), new Rectangle2D.Double(xpos,ypos,50,50 ) );
			xpos += COL_SPACING;
		}
		
		
	}
	
		
	/* used for mapping pairs states to an edge*/
	private class StatePair {
		State s1;

		State s2;

		public StatePair(State s1, State s2) {
			this.s1 = s1;
			this.s2 = s2;

		}

		@Override
		public int hashCode() {
			int hash1 = s1.getName().hashCode();
			int hash2 = s2.getName().hashCode();
			int tmp = 0xFFFF0000 & hash2;

			int result = (hash1 ^ (hash2 << 16)) ^ (tmp >>> 16);

			return result;
		}

		@Override
		public boolean equals(Object obj) {
			StatePair sp = (StatePair) obj;
			return sp.s1.equals(s1) && sp.s2.equals(s2);
		}
	}
	
	

	public void actionPerformed(ActionEvent e) {
		ArrayList<DefaultEdge> nonLoopers = new ArrayList<DefaultEdge>(edges);
		nonLoopers.removeAll(selfLoops);
		
		if(e.getSource() == itemLinesSpline) {
			Map m = new Hashtable();
			
			GraphConstants.setLineStyle(m, GraphConstants.STYLE_BEZIER);
			
			
			
			graph.getGraphLayoutCache().edit(nonLoopers.toArray(), m);
			
		} else if(e.getSource() == itemLinesStraight) {
			Map m = new Hashtable();

			GraphConstants.setLineStyle(m, GraphConstants.STYLE_ORTHOGONAL);
			
			graph.getGraphLayoutCache().edit(nonLoopers.toArray(), m);
			
		} else if(e.getSource() == itemRoutingOn) {
			
			Map m = new Hashtable();
			
			GraphConstants.setRouting(m, GraphConstants.ROUTING_SIMPLE);
			
			graph.getGraphLayoutCache().edit(nonLoopers.toArray(), m);
			itemLinesStraight.setEnabled(true);
			itemLinesSpline.setEnabled(true);
			
		} else if(e.getSource() == itemRoutingOff) {
			Map m = new Hashtable();
			
			GraphConstants.setRouting(m, GraphConstants.ROUTING_DEFAULT);
			
			graph.getGraphLayoutCache().edit(nonLoopers.toArray(), m);
			itemLinesStraight.setEnabled(false);
			itemLinesSpline.setEnabled(false);
		} else if(e.getSource() == itemFileExit) {
			
			System.exit(0);
		}
	}
}

class BFSNode<M> {
	public M value;
	public int depth;
	
	public BFSNode(M value, int depth) {
		this.value = value;
		this.depth = depth;
	}
	
}
