package parser;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.TreeSet;

import core.FiniteStateMachine;
import core.State;
import core.StateType;


public class FSMParser {
	private static final char NEW_TRANSITION_SYMBOL = '+';

	private static final char TRANSITION_SYMBOL = '.';

	private static final char EOL = '\n';
	
	private static final int EOS = -1;

	private static Reader reader;

	private static int col = 0;
	
	private static int line = 0;
	
	private static int ch;

	private static FiniteStateMachine result;
	
	private static State currentState;
	
	private static TreeSet<String> definedStateNames;
	
	public static FiniteStateMachine readFSM(String filename) throws IOException, FSMParseException{
		FileReader fr = new FileReader(filename);
		FiniteStateMachine fsm = FSMParser.parseFSM(fr);
		fr.close();
		return fsm;
	}
	
	public static FiniteStateMachine parseFSM(Reader r) throws FSMParseException, IOException {
		reader = r;
		
		col = 0;
		
		definedStateNames = new TreeSet<String>();
		
		line = 0;
		
		result = new FiniteStateMachine();	
		
		read();
		
		while(ch != EOS) {
			
			
			//Skip blanks before state name, return result on streamend, restart loop on newline
			if(isBlank())		
				skipBlanks();
			if(ch == EOS) {
				return finish();
			}
			if(ch == EOL) {
				read();
				continue;
			}
			
			//read in the name of the state
			String name = readIdentifier();
			//determine whether name is valid, and whether it contains a leading i or f, signifying initial and final state
			currentState = checkAndAddState(name);
			
			
			//Skip blanks before '=', return result on streamend, restart loop on newline
			if(isBlank())		
				skipBlanks();
			if(ch == EOS) {
				return finish();
			}
			if(ch == EOL) {
				read();
				continue;
			}
			
			//expect a '='
			if(ch != '=') {
				throw new FSMParseException("expected a '=', got a '" + (char) ch + "'", line, col);
			}
			
			read();
						
			//Skip blanks before first transition definition, error on streamend, error on newline
			if(isBlank())		
				skipBlanks();
			if(ch == EOS) {
				throw new FSMParseException("unexpected end of stream reached", line, col);
			}
			if(ch == EOL) {
				throw new FSMParseException("unexpected end of line reached", line, col);
			}
			
			//read and add transition
			boolean readAgain = false;
			do {
				skipBlanks();
				
				//recursively read in transition
				topLevelReadTransition(currentState);
				
				skipBlanks();
				
				if(ch==NEW_TRANSITION_SYMBOL) {
					readAgain = true;
					read();
				} else {
					readAgain = false;
				}
				
			} while(readAgain);
			
			if((ch != EOL) && (ch!=EOS)) {
				throw new FSMParseException("Expected '"+ NEW_TRANSITION_SYMBOL + "' or end of line", line, col);
			}
			
			read();
		} 
		
		return finish();
		
	}

	
	//expects to be pointed at first character
	private static void topLevelReadTransition(State baseState) throws FSMParseException, IOException {
		
		State currentBaseState = baseState;
		ArrayList<StateAndSymbol> leaves = null;
		
		do {
			if(leaves != null) {
				State anonState = result.createAnonymousState();
				for(StateAndSymbol leaf : leaves) {
					leaf.state.addTransition(leaf.symbol, anonState);				
				}
				currentBaseState = anonState;
			}
			
			if(ch=='(') {
				
				leaves = recReadTransition(currentBaseState);
							
			} else if(Character.isLowerCase(ch)){
				String ident = readIdentifier();
				leaves = new ArrayList<StateAndSymbol>();
				leaves.add(new StateAndSymbol(ident, currentBaseState));
				
			} else {
				throw new FSMParseException("Unexpected character " + ((char) ch) + " transition term expected", line, col);
			}

			
			
			skipBlanks();
			if(ch!=TRANSITION_SYMBOL) {
				throw new FSMParseException("Unexpected character '" + ((char) ch) + "', '" + TRANSITION_SYMBOL + "' expected ", line, col);
			}
			read();
			skipBlanks();			
		} while(Character.isLowerCase(ch) || (ch=='('));
		
		
		if(!isIdentChar() || Character.isLowerCase((char) ch))
			throw new FSMParseException("Unexpected character '" + ((char) ch) + "', expected a state name with uppercase letter",line,col);
		
		String ident = readIdentifier();
		State succState = getOrCreateState(ident, false, false,false);
		
		for(StateAndSymbol leaf : leaves) {
			leaf.state.addTransition(leaf.symbol, succState);				
		}
		
		
	}
	
	//recursively reads in transition constructs with brackets like "a.(b+c.(d+f.g)).P"
	//Returns all generated states and the symbol with which they perform their transition
	//of the deepest recursion level, e.g. the abovementioned example
	//would ultimately return the anonymous states from which b, d, and g originate.
	//Expects "ch" to be pointed at opening bracket of statement
	private static ArrayList<StateAndSymbol> recReadTransition(State baseState) throws IOException, FSMParseException {
		
		assert(ch=='(');
		

		//Saves the leaf state/symbol of all terms seperated by '+' on this level
		ArrayList<StateAndSymbol> allTermsLeaves = new ArrayList<StateAndSymbol>();
			
		
		//read in a sequence of terms seperated by '+'
		do {
			
			read();
			
			skipBlanks();
			
			//saves the leaf state/symbol of this term
			ArrayList<StateAndSymbol> thisTermLeaves = null;
			
			State currentBaseState = baseState;
					
			boolean readAnotherSequence;
			do {
				//read in a sequence of terms seperated by TRANSITION_SYMBOL ('.')
				if(thisTermLeaves != null) {
					State anonState = result.createAnonymousState();
					for(StateAndSymbol leaf : thisTermLeaves) {
						leaf.state.addTransition(leaf.symbol, anonState);
					}
					
					currentBaseState = anonState;
				}
				
				
				
				
				readAnotherSequence = false;
				
				if(ch=='(') {
					thisTermLeaves = recReadTransition(currentBaseState);
				} else if(Character.isLowerCase(ch)){
					String ident = readIdentifier();
					thisTermLeaves = new ArrayList<StateAndSymbol>();
					thisTermLeaves.add(new StateAndSymbol(ident, currentBaseState));
					
				} else {
					throw new FSMParseException("Unexpected character " + ((char) ch) + " transition term expected", line, col);
				}
				
				
				
				skipBlanks();
				
				if(ch==TRANSITION_SYMBOL) {
					read();
					skipBlanks();
					readAnotherSequence = true;
				} 
								
			} while(readAnotherSequence);
			
			allTermsLeaves.addAll(thisTermLeaves);
			
		} while(ch==NEW_TRANSITION_SYMBOL);
		
		if(ch!=')') {
			throw new FSMParseException("Missing a closing bracket", line, col);
		}
		read();
		
		return allTermsLeaves;
	}
	
	private static class StateAndSymbol {
		public String symbol;
		public State state;
		
		public StateAndSymbol(String symbol, State state) {
			this.symbol = symbol;
			this.state = state;
		}
		
		@Override
		public String toString() {
			StringBuffer name = new StringBuffer(symbol.length() + state.getName().length() + 3);
			return name.append('(').append(state.getName()).append(".").append(symbol).append(')').toString();
		}
	}
	
	private static FiniteStateMachine finish() throws IOException {
		FiniteStateMachine res = result;
		
		reader.close();
		reader = null;
		col = line = ch = 0;
		currentState = null;
		definedStateNames = null;
		result = null;
		
		return res;
	}

	/**
	 * Checks the name of the state and adds it to the resulting fsm  
	 * @param name The name of the state to add
	 * @param definition 
	 * @return The state that was added
	 * @throws FSMParseException
	 */
	private static State checkAndAddState(String name) throws FSMParseException {
		if(name.length()==0) throw new FSMParseException("Name of state expected", line, col);
		int nameStartIndex=0;
		boolean finalState, initialState;
		finalState = initialState = false;
		if(name.charAt(0) == 'i') {
			nameStartIndex = 1;
			initialState = true;
			if((name.length() > 1) && (name.charAt(1) == 'f')) {
				finalState = true;
				nameStartIndex = 2;
			}
			
		} else if(name.charAt(0)=='f') {
			nameStartIndex = 1;
			finalState = true;
			if((name.length() > 1) && (name.charAt(1) == 'i')) {
				initialState = true;
				nameStartIndex = 2;
			}
			
		}
		
		name = name.substring(nameStartIndex);
		if(name.length() == 0) {
			throw new FSMParseException("Name of state expected", line, col);
		}
		if(Character.isLowerCase(name.charAt(0))) {
			throw new FSMParseException("State names may not begin with lower-case letters: " + name, line, col);
		}
		
		
		
		//
		State s = getOrCreateState(name, true, initialState, finalState);
		
		
		
		return s;
	}
	
	/**
	 * Returns the state of the given name if it exists. Otherwise creates a new state.
	 * If the state already existed, if either initiaState or finalState is true
	 * the state's flags are changed correspondingly. If they are false, the state
	 * remains unchanged. 
	 * @param name The name of the state to get or create
	 * @param initialState True if the state is an initial State
	 * @param finalState True if the state is a final state
	 * @param definition true if the state is defined, false if it is only referenced through another state
	 * @return
	 * @throws FSMParseException 
	 */
	private static State getOrCreateState(String name, boolean definition, boolean initialState, boolean finalState) throws FSMParseException {
		State s = result.getState(name);
		
		StateType type;
		
		if(definition) {
			if(definedStateNames.contains(name)) {
				throw new FSMParseException("Redefinition of state " + name, line, col);
			}
			definedStateNames.add(name);
			
			if(initialState && finalState)
				type = StateType.INITIAL_FINAL_STATE;
			else if(initialState)
				type = StateType.INITIAL_STATE;
			else if(finalState)
				type = StateType.FINAL_STATE;
			else type = StateType.NORMAL_STATE;
			
			if(s!=null) {
				//if the state already exists, only change, final and initial status
				result.setType(s, type);
			} else {
				//If it doesn't exist create a new one
				
				s = result.createState(name, type);
			}
	
		} else {
			//not a definition of the state, only a reference
			if(s==null) 
				s=result.createState(name, StateType.NORMAL_STATE);
			
		}
		
		return s;
	}

	
	private static String readIdentifier() throws IOException, FSMParseException {
		if(!isIdentChar()) throw new FSMParseException("unexpected character '"+ (char) ch + "', expected identifier", line, col);
		StringBuffer b = new StringBuffer();		
		do {
			b.append((char) ch);
			read();
		} while(isIdentChar());
			
		return b.toString();
	}
	
	private static void skipBlanks() throws IOException {
		while(isBlank()) {
			read();
		}
	}
	
	private static  boolean isBlank() {
		return (ch==' ') || (ch=='\t');
	}
	
	private static boolean isIdentChar() {
		return Character.isLetterOrDigit(ch)
			|| (ch=='{') || (ch=='}') || (ch=='_')|| (ch==',')|| (ch=='|'); 
	}
	
	private static void read() throws IOException {
		
		try {
			ch = reader.read();
			if(ch!=EOS) col++;
		} catch (IOException e) {
			throw new IOException("Error encountered trying to read stream: " + e.getMessage());
		}
		if ((ch == '\n') || (ch == '\r')) {
			col = 0;
			line++;
			ch = EOL;
		}
		

	}

}
